"""Synchronization primitives."""

import numpy as np
import pytest
from beartype.claw import beartype_package
from beartype.door import is_bearable

beartype_package("abstract_dataloader")

from abstract_dataloader import generic, spec  # noqa: E402


def test_sync_types():
    """Test synchronization protocols."""
    assert is_bearable(generic.Empty(), spec.Synchronization)
    assert is_bearable(generic.Next("dummy"), spec.Synchronization)
    assert is_bearable(generic.Nearest("dummy"), spec.Synchronization)


def _make_timestamps():
    return {
        "sensor1": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float64),
        "sensor2": np.array([0.5, 1.5, 2.5, 3.5, 4.5], dtype=np.float64),
        "sensor3": np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float64),
    }


def test_empty():
    """generic.Empty."""
    ts = _make_timestamps()
    empty = generic.Empty()(ts)
    for k in ts:
        assert empty[k].shape == (0,)

@pytest.mark.parametrize("reference", ["sensor1", "sensor2", "sensor3"])
def test_next(reference):
    """generic.Next."""
    ts = _make_timestamps()
    sync = generic.Next(reference)
    indices = sync(ts)

    assert set(indices.keys()) == set(ts.keys())

    lengths = [len(v) for v in indices.values()]
    assert all(x == lengths[0] for x in lengths)

    for v in indices.values():
        assert np.all(np.diff(v) >= 0)


def test_next_margin():
    """generic.Next, with margin."""
    ts = _make_timestamps()

    sync = generic.Next("sensor1", margin=(1, 1))
    indices = sync(ts)
    assert indices["sensor1"].shape[0] == 1
    assert np.allclose(indices["sensor1"], [2])

    sync2 = generic.Next("sensor1", margin=(0.1, 1.1))
    indices = sync2(ts)
    assert indices["sensor1"].shape[0] == 1
    assert np.allclose(indices["sensor1"], [2])


def test_next_missing():
    """generic.Next, missing reference."""
    sync = generic.Next(reference="sensorX")
    with pytest.raises(KeyError):
        sync(_make_timestamps())


@pytest.mark.parametrize("tol", [0.05, 0.55, np.inf])
def test_nearest(tol):
    """generic.Nearest."""
    ts = _make_timestamps()
    sync = generic.Nearest(reference="sensor3", tol=tol)
    indices = sync(ts)

    assert set(indices.keys()) == set(ts.keys())
    lengths = [len(v) for v in indices.values()]
    assert all(x == lengths[0] for x in lengths)

    if tol < np.inf:
        t_ref = ts["sensor3"]
        for k in indices:
            selected_times = ts[k][indices[k]]
            diff = np.abs(selected_times - t_ref[:len(selected_times)])
            assert np.all(diff < tol)


def test_nearest_missing():
    """generic.Nearest; missing key."""
    sync = generic.Nearest(reference="sensorX", tol=0.05)
    with pytest.raises(KeyError):
        sync(_make_timestamps())


def test_nearest_bounds():
    """generic.Nearest; invalid tol."""
    with pytest.raises(ValueError):
        sync = generic.Nearest(reference="sensorX", tol=-0.5)  # noqa


def test_nearest_margin():
    """generic.Nearest, with margin."""
    ts = _make_timestamps()

    sync = generic.Nearest("sensor1", tol=1.0, margin=(1, 1))
    indices = sync(ts)
    assert indices["sensor1"].shape[0] == 4
    assert np.allclose(indices["sensor1"], [1, 2, 3, 4])

    sync2 = generic.Nearest("sensor1", tol=1.0, margin=(0.1, 1.1))
    indices = sync2(ts)
    assert indices["sensor1"].shape[0] == 3
    assert np.allclose(indices["sensor1"], [1, 2, 3])



"""Type checking the types used for checking."""

from typing import Sequence

import numpy as np
import pytest
from beartype.claw import beartype_package
from beartype.door import is_bearable, is_subhint

beartype_package("abstract_dataloader")

from abstract_dataloader import abstract, generic, spec  # noqa: E402


def test_spec_types():
    """Verify spec inheritance."""
    assert is_subhint(spec.Trace, spec.Dataset)
    assert is_subhint(spec.Sensor, spec.Dataset)
    assert is_subhint(spec.Collate, spec.Transform)


class _ConcreteMetadata(spec.Metadata):

    def __init__(self, n: int = 10):
        self.timestamps = np.arange(n, dtype=np.float64)

class _ConcreteSensor(abstract.Sensor[int, _ConcreteMetadata]):

    def __getitem__(self, index: int | np.integer) -> int:
        return index * 2

class _ConcreteTransform(spec.Transform):
    def __init__(self, n: int = 1):
        self.n = n

    def __call__(self, data):
        return data + self.n

class _ConcretePipeline(abstract.Pipeline[int, int, list[int], list[int]]):
    def collate(self, data: Sequence[int]) -> list[int]:
        return list(data)


def test_abstract_types():
    """Verify abstract class inheritance."""
    meta = _ConcreteMetadata(10)
    sensor = _ConcreteSensor(meta)
    trace = abstract.Trace({"sensor": sensor}, sync=generic.Next("sensor"))
    dataset = abstract.Dataset([trace, trace])
    transform = _ConcreteTransform()
    pipeline = _ConcretePipeline(sample=transform)

    assert is_bearable(meta, spec.Metadata)
    assert is_bearable(sensor, spec.Sensor)
    assert is_bearable(trace, spec.Trace)
    assert is_bearable(dataset, spec.Dataset)
    assert is_bearable(transform, spec.Transform)
    assert is_bearable(pipeline, spec.Pipeline)


@pytest.mark.parametrize("do_sync", [True, False, "manual"])
def test_dataset(do_sync):
    """Test dataset loading."""
    meta = _ConcreteMetadata(10)
    sensor = _ConcreteSensor(meta)

    if do_sync == "manual":
        sync = {"sensor": np.arange(10)}
    elif do_sync:
        sync = generic.Nearest("sensor")
    else:
        sync = None

    trace = abstract.Trace({"sensor": sensor}, sync=sync)
    dataset = abstract.Dataset([trace, trace])

    assert len(sensor) == 10
    assert len(trace) == 10
    assert len(dataset) == 20

    assert sensor.duration == 9.0
    assert sensor.framerate == 1.0

    assert sensor[3] == 6
    assert trace[3] == {'sensor': 6}
    assert dataset[2] == {'sensor': 4}
    assert dataset[13] == {'sensor': 6}
    assert dataset[np.uint8(13)] == {"sensor": 6}
    assert trace['sensor'] is sensor

    for i, j in enumerate(sensor.stream()):
        assert i * 2 == j

    with pytest.raises(IndexError):
        dataset[-1]
    with pytest.raises(IndexError):
        dataset[20]

    acc = []
    for i in sensor.stream(batch=2):
        acc.extend(i)
    assert acc == [i * 2 for i in range(10)]


def test_transform():
    """Test transform composition."""
    transform = _ConcreteTransform()
    pipeline = _ConcretePipeline(
        sample=abstract.Transform(transforms=[transform, transform]),
        batch=lambda data: [x * 2 for x in data])
    raw = [1, 2, 3]
    transformed = [pipeline.sample(x) for x in raw]
    collated = pipeline.collate(transformed)
    processed = pipeline.batch(collated)
    assert transformed == [3, 4, 5]
    assert collated == [3, 4, 5]
    assert processed == [6, 8, 10]

    emptypipeline = _ConcretePipeline()
    raw = [1, 2]
    transformed = [emptypipeline.sample(x) for x in raw]
    collated = emptypipeline.collate(transformed)
    processed = emptypipeline.batch(collated)
    assert transformed == [1, 2]
    assert collated == [1, 2]
    assert processed == [1, 2]



"""Data type system framework following the [ADL recommendations](./types.md).

!!! abstract "Programming Model"

    - Declare types as dataclasses using the provided [`dataclass`][.]
        decorator, which registers them to the global optree namespace.
    - Set each class as `Generic[TArray]`, where [`TArray`][.] is a `TypeVar`
        which is [`ArrayLike`][.], e.g., `torch.Tensor`, `jax.Array` or
        `np.ndarray`.

!!! warning

    This module requires `optree` to be installed.
"""

from dataclasses import field
from typing import (
    Any,
    Protocol,
    TypeVar,
    runtime_checkable,
)

from optree.dataclasses import dataclass as _optree_dataclass
from typing_extensions import dataclass_transform


@dataclass_transform(field_specifiers=(field,))
def dataclass(cls):  # noqa: D103
    """A dataclass decorator which registers into optree's global namespace."""
    return _optree_dataclass(cls, namespace='')


@runtime_checkable
class ArrayLike(Protocol):
    """Array type, e.g., `torch.Tensor | jax.Array | np.ndarray`.

    Use this type to specify arbitrary array types.
    """

    @property
    def shape(self) -> tuple[int, ...]: ...

    @property
    def dtype(self) -> Any: ...


TArray = TypeVar("TArray", bound=ArrayLike)
"""Type variable for [`ArrayLike`][^.] types."""



"""Abstract Dataloader Specifications.

The implementations here provide "duck type"
[protocol](https://peps.python.org/pep-0544/) definitions of key data loading
primitives. In order to implement the specification, users simply need to
"fill in" the methods described here for the types which they wish to
implement.

!!! type-parameters "Type Parameters"

    ADL specification protocol types are defined as
    [generics](https://typing.python.org/en/latest/reference/generics.html),
    which are parameterized by other types. These type parameters are
    documented by a `Type Parameters` section where applicable.

!!! composition-rules "Composition Rules"

    ADL protocols which can be *composed* together are documented by a
    `Composition Rules` section.
"""

from collections.abc import Iterator, Sequence
from typing import Generic, Protocol, overload, runtime_checkable

import numpy as np
from jaxtyping import Float, Integer
from typing_extensions import TypeVar

__all__ = [
    "Dataset", "Metadata", "Sensor", "Synchronization", "Trace", "Pipeline",
    "Transform", "Collate"
]


@runtime_checkable
class Metadata(Protocol):
    """Sensor metadata.

    All sensor metadata is expected to be held in memory during training, so
    great effort should be taken to minimize its memory usage. Any additional
    information which is not strictly necessary for book-keeping, or which
    takes more than negligible space, should be loaded as data instead.

    !!! note

        This can be a `@dataclass`, [`typing.NamedTuple`][typing.NamedTuple],
        or a fully custom type - it just has to expose a `timestamps`
        attribute.

    Attributes:
        timestamps: measurement timestamps, in seconds. Nominally in epoch
            time; must be consistent within each trace (but not necessarily
            across traces). Suggested type: `float64,` which gives precision of
            <1us.
    """

    timestamps: Float[np.ndarray, "N"]


TSample = TypeVar("TSample", covariant=True)
TMetadata = TypeVar("TMetadata", bound=Metadata)


@runtime_checkable
class Sensor(Protocol, Generic[TSample, TMetadata]):
    """A sensor, consisting of a synchronous time-series of measurements.

    This protocol is parameterized by generic `TSample` and `TMetadata` types,
    which can encode the expected data type of this sensor. For example:

    ```python
    class Point2D(TypedDict):
        x: float
        y: float

    def point_transform(point_sensor: Sensor[Point2D, Any]) -> T:
        ...
    ```

    This encodes an argument, `point_sensor`, which expected to be a sensor
    that reads data with type `Point2D`, but does not specify a metadata type.

    Type Parameters:
        - `TSample`: sample data type which this `Sensor` returns. As a
            convention, we suggest returning "batched" data by default, i.e.
            with a leading singleton axis.
        - `TMetadata`: metadata type associated with this sensor; must
            implement [`Metadata`][^.].

    Attributes:
        metadata: sensor metadata, including timestamp information.
    """

    metadata: TMetadata

    def stream(self) -> Iterator[TSample]:
        """Stream values recorded by this sensor.

        Returns:
            An iterator yielding successive samples.
        """
        ...

    def __getitem__(self, index: int | np.integer) -> TSample:
        """Fetch measurements from this sensor, by index.

        Args:
            index: sample index, in the sensor scope.

        Returns:
            Loaded sample.
        """
        ...

    def __len__(self) -> int:
        """Total number of measurements."""
        ...


@runtime_checkable
class Synchronization(Protocol):
    """Synchronization protocol for asynchronous time-series.

    Defines a rule for creating matching sensor index tuples which correspond
    to some kind of global index.

    !!! info "Generic Implementations"

        The following generic implementations are included with
        [`abstract_dataloader.generic`][abstract_dataloader.generic]:

        - [`Empty`][abstract_dataloader.generic.Empty]: a no-op for
            intializing a trace without any synchronization (i.e., just as a
            container of sensors).
        - [`Nearest`][abstract_dataloader.generic.Nearest]: find the nearest
            measurement for each sensor relative to the reference sensor's
            measurements.
        - [`Next`][abstract_dataloader.generic.Next]: find the next
            measurement for each sensor relative to the reference sensor's
            measurements.
    """

    def __call__(
        self, timestamps: dict[str, Float[np.ndarray, "_N"]]
    ) -> dict[str, Integer[np.ndarray, "M"]]:
        """Apply synchronization protocol.

        Args:
            timestamps: sensor timestamps. Each key denotes a different sensor
                name, and the value denotes the timestamps for that sensor.

        Returns:
            A dictionary, where keys correspond to each sensor, and values
                correspond to the indices which map global indices to sensor
                indices, i.e. `global[sensor, i] = sensor[sync[sensor] [i]]`.
        """
        ...


@runtime_checkable
class Trace(Protocol, Generic[TSample]):
    """A sensor trace, consisting of multiple simultaneously-recording sensors.

    This protocol is parameterized by a generic `Sample` type, which can encode
    the expected data type of this trace.

    !!! question "Why not `Sequence`?"

        While collections of simultaneously-recorded sensor data are commonly
        referred to as "sequences," the `Sequence` name conflicts with the
        python standard-library
        [`Sequence`](https://docs.python.org/3/library/collections.abc.html#collections.abc.Sequence)
        type. Since the abstract dataloader heavily references these
        [abstract types][collections.abc] (and you should too!), we use "Trace"
        to avoid conflicts.

    Type Parameters:
        - `Sample`: sample data type which this `Trace` returns. As a
            convention, we suggest returning "batched" data by default, i.e.
            with a leading singleton axis.
    """

    @overload
    def __getitem__(self, index: str) -> Sensor: ...

    @overload
    def __getitem__(self, index: int | np.integer) -> TSample: ...

    def __getitem__(self, index: int | np.integer | str) -> TSample | Sensor:
        """Get item from global index (or fetch a sensor by name).

        !!! info

            For the user's convenience, traces can be indexed by a `str` sensor
            name, returning that [`Sensor`][^^.]. While we are generally wary
            of requiring "quality of life" features, we include this since a
            simple `isinstance(index, str)` check suffices to implement this
            feature.

        Args:
            index: sample index, or sensor name.

        Returns:
            Loaded sample if `index` is an integer type, or the appropriate
            [`Sensor`][^^.] if `index` is a `str`.
        """
        ...

    def __len__(self) -> int:
        """Total number of sensor-tuple samples."""
        ...


@runtime_checkable
class Dataset(Protocol, Generic[TSample]):
    """A dataset, consisting of multiple traces concatenated together.

    !!! note "[`Trace`][^.] subtypes [`Dataset`][^.]"

        Due to the type signatures, a [`Trace`][^.] is actually a subtype of
        [`Dataset`][^.]. This means that a dataset which implements a
        collection of traces can also take a collection of datasets!

    Type Parameters:
        - `TSample`: sample data type which this `Dataset` returns. As a
            convention, we suggest returning "batched" data by default, i.e.
            with a leading singleton axis.
    """

    def __getitem__(self, index: int | np.integer) -> TSample:
        """Fetch item from this dataset by global index.

        Args:
            index: sample index.

        Returns:
            Loaded sample.
        """
        ...

    def __len__(self) -> int:
        """Total number of samples in this dataset."""
        ...


TRaw = TypeVar("TRaw", infer_variance=True)
TTransformed = TypeVar("TTransformed", infer_variance=True)
TCollated = TypeVar("TCollated", infer_variance=True)
TProcessed = TypeVar("TProcessed", infer_variance=True)


@runtime_checkable
class Transform(Protocol, Generic[TRaw, TTransformed]):
    """Sample or batch-wise transform.

    !!! note

        This protocol is a suggestively-named equivalent to
        `Callable[[TRaw], TTransformed]` or `Callable[[Any], Any]`.

    Composition Rules:
        - `Transform` can be freely composed, as long as each transform's
          `TTransformed` matches the next transform's `TRaw`; this composition
          is implemented by [`abstract.Transform`][abstract_dataloader.].
        - Composed `Transform`s result in another `Transform`:
          ```
          Transform[T2, T3] (.) Transform[T1, T2] = Transform[T1, T3].
          ```

    Type Parameters:
        - `TRaw`: Input data type.
        - `TTransformed`: Output data type.
    """

    def __call__(self, data: TRaw) -> TTransformed:
        """Apply transform to a single sample.

        Args:
            data: A single `TRaw` data sample.

        Returns:
            A single `TTransformed` data sample.
        """
        ...


@runtime_checkable
class Collate(Protocol, Generic[TTransformed, TCollated]):
    """Data collation.

    !!! note

        This protocol is a equivalent to
        `Callable[[Sequence[TTransformed]], TCollated]`. `Collate` can also
        be viewed as a special case of `Transform`, where the input type
        `TRaw` must be a `Sequence[...]`.

    Composition Rules:
        - `Collate` can only be composed in parallel, and can never be
          sequentially composed.

    Type Parameters:
        - `TTransformed`: Input data type.
        - `TCollated`: Output data type.
    """

    def __call__(self, data: Sequence[TTransformed]) -> TCollated:
        """Collate a set of samples.

        Args:
            data: A set of `TTransformed` samples.

        Returns:
            A `TCollated` batch.
        """
        ...


@runtime_checkable
class Pipeline(
    Protocol, Generic[TRaw, TTransformed, TCollated, TProcessed]
):
    """Dataloader transform pipeline.

    This protocol is parameterized by four type variables which encode the
    different data formats at each stage in the pipeline. This forms a
    `Raw -> Transformed -> Collated -> Processed` pipeline with three
    transforms:

    - [`sample`][.]: a sample to sample transform; can be sequentially
      assembled from one or more [`Transform`][^.]s.
    - [`collate`][.]: a list-of-samples to batch transform. Can use exactly one
      [`Collate`][^.].
    - [`batch`][.]: a batch to batch transform; can be sequentially assembled
      from one or more [`Transform`][^.]s.

    Composition Rules:
        - A full `Pipeline` can be sequentially pre-composed and/or
          post-composed with one or more [`Transform`][^.]s; this is
          implemented by [`generic.ComposedPipeline`][abstract_dataloader.].
        - `Pipeline`s can always be composed in parallel; this is implemented
          by [`generic.ParallelPipelines`][abstract_dataloader.], with a
          pytorch [`nn.Module`][torch.]-compatible version in
          [`torch.ParallelPipelines`][abstract_dataloader.].

    Type Parameters:
        - `TRaw`: Input data format.
        - `TTransformed`: Data after the first `transform` step.
        - `TCollated`: Data after the second `collate` step.
        - `TProcessed`: Output data format.
    """

    def sample(self, data: TRaw) -> TTransformed:
        """Transform single samples.

        - Operates on single samples, nominally on the CPU-side of a
          dataloader.
        - This method is both sequentially and parallel composable.

        Args:
            data: A single `TRaw` data sample.

        Returns:
            A single `TTransformed` data sample.
        """
        ...

    def collate(self, data: Sequence[TTransformed]) -> TCollated:
        """Collate a list of data samples into a GPU-ready batch.

        - Operates on the CPU-side of the dataloader, and is responsible for
          aggregating individual samples into a batch (but not transferring to
          the GPU).
        - Analogous to the `collate_fn` of a
          [pytorch dataloader](https://pytorch.org/docs/stable/data.html).
        - This method is not sequentially composable.

        Args:
            data: A sequence of `TTransformed` data samples.

        Returns:
            A `TCollated` collection of the input sequence.
        """
        ...

    def batch(self, data: TCollated) -> TProcessed:
        """Transform data batch.

        - Operates on a batch of data, nominally on the GPU-side of a
          dataloader.
        - This method is both sequentially and parallel composable.

        !!! info "Implementation as `torch.nn.Module`"

            If these `Transforms` require GPU state, it may be helpful to
            implement it as a `torch.nn.Module`. In this case, `batch`
            should redirect to `__call__`, which in turn redirects to
            [`nn.Module.forward`][torch.] in order to handle any registered
            pytorch hooks.

        Args:
            data: A `TCollated` batch of data, nominally already sent to the
                GPU.

        Returns:
            The `TProcessed` output, ready for the downstream model.
        """
        ...



"""Pytorch-specific components."""

from collections.abc import Sequence
from typing import Any, Generic, Literal, TypeVar, cast

import numpy as np
import torch
from torch.utils.data import Dataset

from abstract_dataloader import spec

TRaw = TypeVar("TRaw")
TTransformed = TypeVar("TTransformed")
TCollated = TypeVar("TCollated")
TProcessed = TypeVar("TProcessed")


def _get_treelib():
    """Get a tree manipulation library."""
    try:
        # Don't give type checking errors since this is optional
        import optree  # type: ignore
        return optree
    # No coverage and no testing on optree-excluding import.
    # If you run into problems, please just install optree.
    except ImportError:  # pragma: no cover
        try:
            return torch.utils._pytree  # type: ignore
        except AttributeError:
            raise NotImplementedError(
                "No tree_map implementation found: `optree` is not "
                "installed, and the pytorch `_pytree` utility is not "
                "present.")


class TransformedDataset(Dataset[TTransformed], Generic[TRaw, TTransformed]):
    """Pytorch-compatible dataset with transformation applied.

    Extends [`torch.utils.data.Dataset`][?torch.utils.data.Dataset],
    implementing a torch "map-style" dataset.

    Type Parameters:
        - `TRaw`: raw data type from the dataloader.
        - `TTransformed`: output data type from the provided transform function.

    Args:
        dataset: source dataset.
        transform: transformation to apply to each sample when loading (note
            that `Transform[TRaw, TTransformed]` is equivalent to
            `Callable[[TRaw], TTransformed]`).
    """

    def __init__(
        self, dataset: spec.Dataset[TRaw],
        transform: spec.Transform[TRaw, TTransformed]
    ) -> None:
        self.dataset = dataset
        self.transform = transform

    def __getitem__(self, index: int | np.integer) -> TTransformed:
        """Map-style dataset indexing.

        Args:
            index: dataset index; passthrough to the underlying `Dataset`.

        Returns:
            Transformed sample.
        """
        return self.transform(self.dataset[index])

    def __len__(self) -> int:
        """Dataset length; passthrough to the underlying `Dataset`."""
        return len(self.dataset)

    def __repr__(self) -> str:
        """Friendly name."""
        return f"Transformed({repr(self.dataset)})"


class StackedSequencePipeline(
    torch.nn.Module,
    spec.Pipeline[
        Sequence[TRaw], Sequence[TTransformed], TCollated, TProcessed]
):
    """Modify a pipeline to act on sequences.

    Unlike the generic [`generic.SequencePipeline`][abstract_dataloader.]
    implementation, this class places the sequence axis directly inside each
    tensor, so that each data type has axes `(batch, sequence, ...)`. For the
    same input,

    ```
    [
        [Raw[s=0, t=0], Raw[s=0, t=1], ... Raw[s=0, t=n]]
        [Raw[s=1, t=0], Raw[s=1, t=1], ... Raw[s=1, t=n]]
        ...
        [Raw[s=b, t=0], Raw[s=b, t=1], ... Raw[s=b, t=n]
    ]
    ```

    this pipeline instead yields

    ```python
    Processed[s=0...b] [t=0...n].
    ```

    !!! info

        This class requires that all outputs of `.collate()` are pytorch
        tensors. Furthermore, batches must be treated as an additional leading
        axis by both `.collate` and `.forward`.

    !!! warning

        Since the output has an additional axis, it does not necessarily have
        the same type as the underlying transform!

    This is accomplished by appropriately reshaping the data to use the
    batch-vectorized underlying implementation:

    - `.sample`: apply the pipeline to each sample across the additional
      sequence axis.
    - `.collate`: concatenate all sequences into a single `list[Raw]`, instead
      of a `list[list[Raw]]`. Then, collate the list, and reshape back into
      `batch sequence ...` order.
    - `.batch`: flatten the collated data back to a `(batch sequence) ...`
      single leading batch axis, apply the pipeline, and reshape back.

    Type Parameters:
        - `PRaw`, `PTransformed`, `PCollated`, `PProcessed`: see
          [`Pipeline`][abstract_dataloader.spec.].

    Args:
        pipeline: pipeline to transform to accept sequences.
    """

    def __init__(
        self, pipeline: spec.Pipeline[
            TRaw, TTransformed, TCollated, TProcessed]
    ) -> None:
        super().__init__()
        self.pipeline = pipeline
        self.treelib = _get_treelib()

    def sample(self, data: Sequence[TRaw]) -> list[TTransformed]:
        return [self.pipeline.sample(x) for x in data]

    def collate(self, data: Sequence[Sequence[TTransformed]]) -> Any:
        data_flat = sum((list(x) for x in data), start=[])
        collated_flat = self.pipeline.collate(data_flat)
        unflattened = self.treelib.tree_map(
            lambda x: x.reshape(len(data), -1, *x.shape[1:]),
            collated_flat)   # type: ignore
        return unflattened

    def batch(self, data: Any) -> Any:
        batch = self.treelib.tree_leaves(data)[0].shape[0]  # type: ignore
        flattened = self.treelib.tree_map(
            lambda x: x.reshape(-1, *x.shape[2:]), data)
        transformed = self.pipeline.batch(cast(TCollated, flattened))
        unflattened = self.treelib.tree_map(
            lambda x: x.reshape(batch, -1, *x.shape[1:]),
            transformed)  # type: ignore
        return unflattened


class Collate(spec.Collate[TTransformed, TCollated]):
    """Generic numpy to pytorch collation.

    Converts numpy arrays to pytorch tensors, and either stacks or concatenates
    each value.

    Type Parameters:
        - `TTransformed`: input sample type.
        - `TCollated`: output collated type.

    Args:
        mode: whether to `stack` or `concat` during collation.
    """

    def __init__(self, mode: Literal["stack", "concat"] = "concat") -> None:
        self.mode = mode
        self.treelib = _get_treelib()

    def __call__(self, data: Sequence[TTransformed]) -> TCollated:
        if self.mode == "concat":
            return self.treelib.tree_map(
                lambda *x: torch.concat([torch.from_numpy(s) for s in x]),
                *data)  # type: ignore
        else:
            return self.treelib.tree_map(
                lambda *x: torch.stack([torch.from_numpy(s) for s in x]),
                *data)  # type: ignore



"""Pytorch versions of generic components."""

from typing import Any, TypeVar, cast

import torch

from abstract_dataloader import generic, spec

PRaw = TypeVar("PRaw", bound=dict[str, Any])
PTransformed = TypeVar("PTransformed", bound=dict[str, Any])
PCollated = TypeVar("PCollated", bound=dict[str, Any])
PProcessed = TypeVar("PProcessed", bound=dict[str, Any])


class ParallelTransforms(torch.nn.Module, spec.Transform[PRaw, PTransformed]):
    """Compose multiple transforms, similar to [`ParallelPipelines`][^.].

    Type Parameters:
        - `PRaw`, `PTransformed`, [`Transform`][abstract_dataloader.spec.].

    Args:
        transforms: transforms to compose. The key indicates the subkey to
            apply each transform to.
    """

    def __init__(self, **transforms: spec.Transform) -> None:
        super().__init__()
        self.transforms = transforms
        self._transforms = torch.nn.ModuleDict({
            k: v for k, v in transforms.items()
            if isinstance(v, torch.nn.Module)})

    def __call__(self, data: PRaw) -> PTransformed:
        return cast(
            PTransformed,
            {k: v(data[k]) for k, v in self.transforms.items()})


class ParallelPipelines(
    torch.nn.Module,
    generic.ParallelPipelines[PRaw, PTransformed, PCollated, PProcessed]
):
    """Transform Compositions, modified for Pytorch compatibility.

    Any [`nn.Module`][torch.] transforms are registered to a separate
    [`nn.ModuleDict`][torch.]; the original `.transforms` attribute is
    maintained with references to the full pipeline.

    See [`generic.ParallelPipelines`][abstract_dataloader.]
    for more details about this implementation. `.forward` and `.__call__`
    should work as expected within pytorch.

    Type Parameters:
        - `PRaw`, `PTransformed`, `PCollated`, `PProcessed`: see
          [`Pipeline`][abstract_dataloader.spec.].

    Args:
        transforms: pipelines to compose. The key indicates the subkey to
            apply each transform to.
    """

    def __init__(self, **transforms: spec.Pipeline) -> None:
        super().__init__()
        self.transforms = transforms
        self._transforms = torch.nn.ModuleDict({
            k: v for k, v in transforms.items()
            if isinstance(v, torch.nn.Module)})

    def forward(self, data: PCollated) -> PProcessed:
        # We have to redefine this for some reason to make torch happy.
        # I think `nn.Module` has a generic `forward` implementation which
        # is clobbering `ComposeTransform`.
        return cast(
            PProcessed,
            {k: v.batch(data[k]) for k, v in self.transforms.items()})

    def batch(self, data: PCollated) -> PProcessed:
        """Alias `batch` to `__call__` to `forward` via `nn.Module`."""
        return self(data)


TRawInner = TypeVar("TRawInner")
TRaw = TypeVar("TRaw")
TTransformed = TypeVar("TTransformed")
TCollated = TypeVar("TCollated")
TProcessed = TypeVar("TProcessed")
TProcessedInner = TypeVar("TProcessedInner")

class ComposedPipeline(
    torch.nn.Module,
    generic.ComposedPipeline[
        TRaw, TRawInner, TTransformed, TCollated, TProcessedInner, TProcessed]
):
    """Compose pipeline sequentially with pre and post transforms.

    Type Parameters:
        - `TRaw`: initial input type.
        - `TRawInner`: output of the pre-composed transform, and input to the
            provided [`Pipeline`][abstract_dataloader.spec].
        - `TCollated`, `TProcessed`: intermediate values for the provided
            [`Pipeline`][abstract_dataloader.spec].
        - `TProcessedInner`: output of the transforms, and input to the
            post-composed transform.
        - `TProcessed`: output type.

    Args:
        pipeline: pipeline to compose.
        pre: pre-transform to apply on the CPU side; skipped if `None`.
        post: post-transform to apply on the GPU side; skipped if `None`.
    """

    def __init__(
        self, pipeline: spec.Pipeline[
            TRawInner, TTransformed, TCollated, TProcessedInner],
        pre: spec.Transform[TRaw, TRawInner] | None = None,
        post: spec.Transform[TProcessedInner, TProcessed] | None = None
    ) -> None:
        super().__init__()
        self.pipeline = pipeline
        self.pre = pre
        self.post = post

        self.collate = pipeline.collate


class SequencePipeline(
    torch.nn.Module,
    generic.SequencePipeline[TRaw, TTransformed, TCollated, TProcessed]
):
    """Transform which passes an additional sequence axis through.

    The given `Pipeline` is modified to accept `Sequence[...]` for each
    data type in its pipeline, and return a `list[...]` across the additional
    axis, thus "passing through" the axis.

    For example, suppose a sequence dataloader reads

    ```
    [
        [Raw[s=0, t=0], Raw[s=0, t=1], ... Raw[s=0, t=n]]
        [Raw[s=1, t=0], Raw[s=1, t=1], ... Raw[s=1, t=n]]
        ...
        [Raw[s=b, t=0], Raw[s=b, t=1], ... Raw[s=b, t=n]
    ]
    ```

    for sequence length `t = 0...n` and batch sample `s = 0...b`. For sequence
    length `t`, the output of the transforms will be batched with the sequence
    on the outside:

    ```
    [
        Processed[s=0...b] [t=0],
        Processed[s=0...b] [t=1],
        ...
        Processed[s=0...b] [t=n]
    ]
    ```

    Type Parameters:
        - `TRaw`, `TTransformed`, `TCollated`, `TProcessed`: see
          [`Pipeline`][abstract_dataloader.spec.].

    Args:
        pipeline: input pipeline.
    """

    def __init__(
        self, pipeline: spec.Pipeline[
            TRaw, TTransformed, TCollated, TProcessed]
    ) -> None:
        super().__init__()
        self.pipeline = pipeline



"""Pytorch-ADL wrappers.

These implementations are a superset of the [`generic`][abstract_dataloader.]
components, and provide interoperability with pytorch dataloaders, modules,
etc. For example, any [`Pipeline`][abstract_dataloader.spec.]-related
components which could contain pytorch [`nn.Module`][torch.nn.Module]s are
modified to subclass `nn.Module` in order to properly register them.

!!! warning

    This module is not automatically imported; you will need to explicitly
    import it:

    ```python
    from abstract_dataloader import torch as adl_torch
    ```

    Since pytorch is not declared as a required dependency, you will also need
    to install `torch` (or install the `torch` extra with
    `pip install abstract_dataloader[torch]`).

!!! note

    Recursive tree operations such as reshaping and stacking are performed
    using the `optree` library, or, if that is not present,
    `torch.utils._pytree`, which implements equivalent functionality. If
    `torch.utils._pytree` is removed in a later version, the constructor will
    raise `NotImplementedError`, and this fallback will need to be replaced.

!!! warning

    Custom data container classes such as `@dataclass` are only supported if
    `optree` is installed, and they are [registered with optree](
    https://optree.readthedocs.io/en/stable/dataclasses.html). However, `dict`,
    `list`, `tuple`, and equivalent types such as `TypedDict` and `NamedTuple`
    will work [out of the box](types.md).
"""

# Reimports to expose the same interface as `generic`.
from abstract_dataloader.generic import Empty, Metadata, Nearest, Next, Window

from .generic import (
    ComposedPipeline,
    ParallelPipelines,
    ParallelTransforms,
    SequencePipeline,
)
from .torch import Collate, StackedSequencePipeline, TransformedDataset

__all__ = [
    "ComposedPipeline", "ParallelPipelines", "ParallelTransforms",
    "Metadata", "SequencePipeline", "Window",
    "Empty", "Nearest", "Next",
    # Pytorch-specific Components
    "StackedSequencePipeline", "TransformedDataset", "Collate"
]



"""Objective base classes and specifications.

!!! abstract "Programming Model"

    - An [`Objective`][.] is a callable which returns a (batched) scalar loss
      and a dictionary of metrics.
    - Objectives can be combined into a higher-order objective,
      [`MultiObjective`][.], which combines their losses and aggregates their
      metrics; specify these objectives using a [`MultiObjectiveSpec`][.].
"""

from abc import abstractmethod
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass, field
from typing import Any, Generic, Protocol, TypeVar, cast, runtime_checkable

import numpy as np
import wadler_lindig as wl
from jaxtyping import Float, Shaped, UInt8
from typing_extensions import TypeVar

from .types import TArray

YTrue = TypeVar("YTrue", infer_variance=True)
YPred = TypeVar("YPred", infer_variance=True)


@dataclass(frozen=True)
class VisualizationConfig:
    """General-purpose visualization configuration.

    Objectives which make use of this configuration may ignore the provided
    values.

    Attributes:
        cols: number of columns to tile images for in-training visualizations.
        width: width of each sample when rendered.
        height: height of each sample when rendered.
        cmaps: colormaps to use, where values correspond to the name of a
            matplotlib colormap or a numpy array of enumerated RGB values.
    """

    cols: int = 8
    width: int = 512
    height: int = 256
    cmaps: Mapping[
        str, str | UInt8[np.ndarray, "N 3"]] = field(default_factory=dict)


@runtime_checkable
class Objective(Protocol, Generic[TArray, YTrue, YPred]):
    """Composable training objective.

    !!! note

        Metrics should use `torch.no_grad()` to make sure gradients are not
        computed for non-loss metrics!

    Type Parameters:
        - `TArray`: backend (`jax.Array`, `torch.Tensor`, etc.)
        - `YTrue`: ground truth data type.
        - `YPred`: model output data type.
    """

    @abstractmethod
    def __call__(
        self, y_true: YTrue, y_pred: YPred, train: bool = True
    ) -> tuple[Float[TArray, "batch"], dict[str, Float[TArray, "batch"]]]:
        """Training metrics implementation.

        Args:
            y_true: data channels (i.e. dataloader output).
            y_pred: model outputs.
            train: Whether in training mode (i.e. skip expensive metrics).

        Returns:
            A tuple containing the loss and a dict of metric values.
        """
        ...

    def visualizations(
        self, y_true: YTrue, y_pred: YPred
    ) -> dict[str, UInt8[np.ndarray, "H W 3"]]:
        """Generate visualizations for each entry in a batch.

        This method may return an empty dict.

        !!! note

            This method should be called only from a "detached" CPU thread so
            as not to affect training throughput; the caller is responsible for
            detaching gradients and sending the data to the CPU. As such,
            implementations are free to use CPU-specific methods.

        Args:
            y_true: data channels (i.e., dataloader output).
            y_pred: model outputs.

        Returns:
            A dict, where each key is the name of a visualization, and the
                value is a stack of RGB images in HWC order, detached from
                Torch and sent to a numpy array.
        """
        ...

    def render(
        self, y_true: YTrue, y_pred: YPred, render_gt: bool = False
    ) -> dict[str, Shaped[np.ndarray, "batch ..."]]:
        """Render model outputs and/or ground truth for later analysis.

        This method may return an empty dict.

        ??? question "How does this differ from `visualizations`?"

            Unlike `visualizations`, which is expected to return a single
            RGB image per batch, `render` is:

            - expected to return a unique rendered value per sample, and
            - may have arbitrary types (as long as they are a numpy arrays).

        Args:
            y_true: data channels (i.e. dataloader output).
            y_pred: model outputs.
            render_gt: whether to render ground truth data.

        Returns:
            A dict, where each key is the name of a rendered output, and the
                value is a numpy array of the rendered data (e.g., an image).
        """
        ...


YTrueAll = TypeVar("YTrueAll", infer_variance=True)
YPredAll = TypeVar("YPredAll", infer_variance=True)


@dataclass
class MultiObjectiveSpec(Generic[YTrue, YPred, YTrueAll, YPredAll]):
    """Specification for a single objective in a multi-objective setup.

    The inputs and outputs for each objective are specified using `y_true` and
    `y_pred`:

    - `None`: The provided `y_true` and `y_pred` are passed directly to the
        objective. This means that if multiple objectives all use `None`, they
        will all receive the same data that comes from the dataloader.
    - `str`: The key indexes into a mapping which has the `y_true`/`y_pred` key,
        or an object which has a matching attribute.
    - `Sequence[str]`: Each key indexes into the layers of a nested mapping or
        object.
    - `Callable`: The callable is applied to the provided `y_true` and `y_pred`.

    !!! warning

        The user is responsible for ensuring that the `y_true` and `y_pred`
        keys or callables index the appropriate types for this objective.

    Type Parameters:
        - `YTrue`: objective ground truth data type.
        - `YHat`: objective model prediction data type.
        - `YTrueAll`: type of all ground truth data (as loaded by the
            dataloader).
        - `YHatAll`: type of all model output data (as produced by the model).

    Attributes:
        objective: The objective to use.
        weight: Weight of the objective in the overall loss.
        y_true: Key or callable to index into the ground truth data.
        y_pred: Key or callable to index into the model output data.
    """

    objective: Objective
    weight: float = 1.0
    y_true: str | Sequence[str] | Callable[[YTrueAll], YTrue] | None = None
    y_pred: str | Sequence[str] | Callable[[YPredAll], YPred] | None = None

    def _index(
        self, data: Any, key: str | Sequence[str] | Callable | None
    ) -> Any:
        """Index into data using the key or callable."""
        def dereference(obj, k):
            if isinstance(obj, Mapping):
                try:
                    return obj[k]
                except KeyError as e:
                    raise KeyError(
                        f"Key {k} not found: {wl.pformat(obj)}") from e
            else:
                try:
                    return getattr(obj, k)
                except AttributeError as e:
                    raise AttributeError(
                        f"Attribute {k} not found: {wl.pformat(obj)}") from e

        if isinstance(key, str):
            return dereference(data, key)
        elif isinstance(key, Sequence):
            for k in key:
                data = dereference(data, k)
            return data
        elif callable(key):
            return key(data)
        else:   # key is None
            return data

    def index_y_true(self, y_true: YTrueAll) -> YTrue:
        """Get indexed ground truth data.

        Args:
            y_true: All ground truth data (as loaded by the dataloader).

        Returns:
            Indexed ground truth data.
        """
        return self._index(y_true, self.y_true)

    def index_y_pred(self, y_pred: YPredAll) -> YPred:
        """Get indexed model output data.

        Args:
            y_pred: All model output data (as produced by the model).

        Returns:
            Indexed model output data.
        """
        return self._index(y_pred, self.y_pred)


class MultiObjective(Objective[TArray, YTrue, YPred]):
    """Composite objective that combines multiple objectives.

    ??? example "Hydra Configuration"

        If using [Hydra](https://hydra.cc/docs/intro/) for dependency
        injection, a `MultiObjective` configuration should look like this:
        ```yaml
        objectives:
        name:
            objective:
                _target_: ...
                kwargs: ...
            weight: 1.0
            y_true: "y_true_key"
            y_pred: "y_pred_key"
        ...
        ```

    Type Parameters:
        - `YTrue`: ground truth data type.
        - `YHat`: model output data type.

    Args:
        objectives: multiple objectives, organized by name; see
            [`MultiObjectiveSpec`][^.]. Each objective can also be provided as
            a dict, in which case the key/values are passed to
            `MultiObjectiveSpec`.
    """

    def __init__(self, **objectives: Mapping | MultiObjectiveSpec) -> None:
        if len(objectives) == 0:
            raise ValueError("At least one objective must be provided.")

        self.objectives = {
            k: v if isinstance(v, MultiObjectiveSpec)
            else MultiObjectiveSpec(**v)
            for k, v in objectives.items()}

    def __call__(
        self, y_true: YTrue, y_pred: YPred, train: bool = True
    ) -> tuple[Float[TArray, "batch"], dict[str, Float[TArray, "batch"]]]:
        loss = 0.
        metrics = {}
        for k, v in self.objectives.items():
            k_loss, k_metrics = v.objective(
                v.index_y_true(y_true), v.index_y_pred(y_pred), train=train)
            loss += k_loss * v.weight

            for name, value in k_metrics.items():
                metrics[f"{k}/{name}"] = value

        # We assure that there's at least one objective.
        loss = cast(Float[TArray, ""] | Float[TArray, "batch"], loss)
        return loss, metrics

    def visualizations(
        self, y_true: YTrue, y_pred: YPred
    ) -> dict[str, UInt8[np.ndarray, "H W 3"]]:
        images = {}
        for k, v in self.objectives.items():
            k_images = v.objective.visualizations(
                v.index_y_true(y_true), v.index_y_pred(y_pred))
            for name, image in k_images.items():
                images[f"{k}/{name}"] = image
        return images

    def render(
        self, y_true: YTrue, y_pred: YPred, render_gt: bool = False
    ) -> dict[str, Shaped[np.ndarray, "batch ..."]]:
        rendered = {}
        for k, v in self.objectives.items():
            k_rendered = v.objective.render(
                v.index_y_true(y_true), v.index_y_pred(y_pred),
                render_gt=render_gt)
            for name, image in k_rendered.items():
                rendered[f"{k}/{name}"] = image
        return rendered



"""Generic Time Synchronization Protocols."""

import numpy as np
from jaxtyping import Float64, UInt32

from abstract_dataloader import spec


class Empty(spec.Synchronization):
    """Dummy synchronization which does not synchronize sensor pairs.

    No samples will be registered, and the trace can only be used as a
    collection of sensors.
    """

    def __call__(
        self, timestamps: dict[str, Float64[np.ndarray, "_N"]]
    ) -> dict[str, UInt32[np.ndarray, "M"]]:
        """Apply synchronization protocol.

        Args:
            timestamps: input sensor timestamps.

        Returns:
            Synchronized index map.
        """
        return {k: np.array([], dtype=np.uint32) for k in timestamps}


class Next(spec.Synchronization):
    """Next sample synchronization, with respect to a reference sensor.

    Applies the following:

    - Find the start time, defined by the earliest time which is observed by
      all sensors, and the end time, defined by the last time which is observed
      by all sensors.
    - Truncate the reference sensor's timestamps to this start and end time,
      and use this as the query timestamps.
    - For each time in the query, find the first sample from each sensor which
      is after this time.

    See [`Synchronization`][abstract_dataloader.spec.] for protocol details.

    Args:
        reference: reference sensor to synchronize to.
        margin: time margin (in seconds; `float`) or index margin
            (in samples; `int`) to apply to the start and end, excluding
            samples within this margin.
    """

    def __init__(
        self, reference: str,
        margin: tuple[int | float, int | float] = (0, 0)
    ) -> None:
        self.reference = reference
        self.margin = margin

    def __call__(
        self, timestamps: dict[str, Float64[np.ndarray, "_N"]]
    ) -> dict[str, UInt32[np.ndarray, "M"]]:
        """Apply synchronization protocol.

        Args:
            timestamps: input sensor timestamps.

        Returns:
            Synchronized index map.
        """
        try:
            ref_time_all = timestamps[self.reference]
        except KeyError:
            raise KeyError(
                f"Reference sensor {self.reference} was not provided in "
                f"timestamps, with keys: {list(timestamps.keys())}")

        start_time = max(t[0] for t in timestamps.values())
        end_time = min(t[-1] for t in timestamps.values())

        if isinstance(self.margin[0], float):
            start_time += self.margin[0]
        if isinstance(self.margin[1], float):
            end_time -= self.margin[1]

        start_idx = np.searchsorted(ref_time_all, start_time)
        end_idx = np.searchsorted(ref_time_all, end_time)

        if isinstance(self.margin[0], int):
            start_idx += self.margin[0]
        if isinstance(self.margin[1], int):
            end_idx -= self.margin[1]

        ref_time = ref_time_all[start_idx:end_idx]
        return {
            k: np.searchsorted(v, ref_time).astype(np.uint32)
            for k, v in timestamps.items()}


class Nearest(spec.Synchronization):
    """Nearest sample synchronization, with respect to a reference sensor.

    Applies the following:

    - Compute the midpoints between observations between each sensor.
    - Find which bin the reference sensor timestamps fall into.
    - Calculate the resulting time delta between timestamps. If this exceeds
      `tol` for any sensor-reference pair, remove this match.

    See [`Synchronization`][abstract_dataloader.spec.] for protocol details.

    Args:
        reference: reference sensor to synchronize to.
        tol: synchronization time tolerance, in seconds. Setting `tol = np.inf`
            works to disable this check altogether.
        margin: time margin (in seconds; `float`) or index margin
            (in samples; `int`) to apply to the start and end time relative to
            the reference sensor, excluding samples within this margin.
    """

    def __init__(
        self, reference: str, tol: float = 0.1,
        margin: tuple[int | float, int | float] = (0, 0)
    ) -> None:
        if tol < 0:
            raise ValueError(
                f"Synchronization tolerance must be positive: {tol} < 0")

        self.tol = tol
        self.reference = reference
        self.margin = margin

    def __call__(
        self, timestamps: dict[str, Float64[np.ndarray, "_N"]]
    ) -> dict[str, UInt32[np.ndarray, "M"]]:
        """Apply synchronization protocol.

        Args:
            timestamps: input sensor timestamps.

        Returns:
            Synchronized index map.
        """
        try:
            t_ref = timestamps[self.reference]
        except KeyError:
            raise KeyError(
                f"Reference sensor {self.reference} was not provided in "
                f"timestamps, with keys: {list(timestamps.keys())}")

        if isinstance(self.margin[0], float):
            t_ref = t_ref[np.argmax(t_ref > t_ref[0] + self.margin[0]):]
        elif isinstance(self.margin[0], int) and self.margin[0] > 0:
            t_ref = t_ref[self.margin[0]:]

        if isinstance(self.margin[1], float):
            t_ref = t_ref[
                :-np.argmax((t_ref < t_ref[-1] - self.margin[1])[::-1])]
        elif isinstance(self.margin[1], int) and self.margin[1] > 0:
            t_ref = t_ref[:-self.margin[1]]

        indices = {
            k: np.searchsorted(
                (t_sensor[:-1] + t_sensor[1:]) / 2, t_ref
            ).astype(np.uint32)
            for k, t_sensor in timestamps.items()}

        valid = np.all(np.array([
           np.abs(timestamps[k][i_nearest] - t_ref) < self.tol
        for k, i_nearest in indices.items()]), axis=0)

        return {k: v[valid] for k, v in indices.items()}



"""Protocol transforms for working with sequences."""

from collections.abc import Callable, Sequence
from dataclasses import dataclass
from multiprocessing.pool import ThreadPool
from typing import Generic, TypeVar, cast

import numpy as np
from jaxtyping import Float64

from abstract_dataloader import abstract, spec

Sample = TypeVar("Sample")
SampleStack = TypeVar("SampleStack")
TMetadata = TypeVar("TMetadata", bound=spec.Metadata)


@dataclass
class Metadata(spec.Metadata):
    """Generic metadata with timestamps.

    Attributes:
        timestamps: epoch timestamps.
    """

    timestamps: Float64[np.ndarray, "N"]


class Window(
    abstract.Sensor[SampleStack, Metadata],
    Generic[SampleStack, Sample, TMetadata]
):
    """Load sensor data across a time window using a sensor transform.

    Use this class as a generic transform to give time history to any sensor:

    ```python
    sensor =  ... # implements spec.Sensor
    with_history = generic.Window(sensor, past=5, future=1, parallel=7)
    ```

    In this example, 5 past samples, the current sample, and 1 future sample
    are loaded on every index:

    ```python
    with_history[i] = [
        sensor[i], sensor[i + 1], ... sensor[i + 5], sensor[i + 6]]
                                            ^
                                # timestamp for synchronization
    ```

    Type Parameters:
        - `SampleStack`: a collated series of consecutive samples. Can simply be
            `list[Sample]`.
        - `Sample`: single observation sample type.
        - `TMetadata`: metadata type for the underlying sensor. Note that the
            `Window` wrapper doesn't actually have metadata type `TMetadata`;
            this type is just passed through from the sensor which is wrapped.

    Args:
        sensor: sensor to wrap.
        collate_fn: collate function for aggregating a list of samples; if not
            specified, the samples are simply returned as a list.
        past: number of past samples, in addition to the current sample. Set
            to `0` to disable.
        future: number of future samples, in addition to the current sample.
            Set to `0` to disable.
        crop: if `True`, crop the first `past` and last `future` samples in
            the reported metadata to ensure that all samples are fully valid.
        parallel: maximum number of samples to load in parallel; if `None`, all
            samples are loaded sequentially.
    """

    def __init__(
        self, sensor: spec.Sensor[Sample, TMetadata],
        collate_fn: Callable[[list[Sample]], SampleStack] | None = None,
        past: int = 0, future: int = 0, crop: bool = True,
        parallel: int | None = None
    ) -> None:
        self.sensor = sensor
        self.past = past
        self.future = future
        self.parallel = parallel

        if collate_fn is None:
            collate_fn = cast(
                Callable[[list[Sample]], SampleStack], lambda x: x)
        self.collate_fn = collate_fn

        self.cropped = crop
        if crop:
            # hack for negative indexing
            _future = None if future == 0 else -future
            self.metadata = Metadata(
                timestamps=sensor.metadata.timestamps[past:_future])
        else:
            self.metadata = Metadata(timestamps=sensor.metadata.timestamps)

    @classmethod
    def from_partial_sensor(
        cls, sensor: Callable[[str], spec.Sensor[Sample, TMetadata]],
        collate_fn: Callable[[list[Sample]], SampleStack] | None = None,
        past: int = 0, future: int = 0, crop: bool = True,
        parallel: int | None = None
    ) -> Callable[[str], "Window[SampleStack, Sample, TMetadata]"]:
        """Partially initialize from partially initialized sensor.

        Use this to create windowed sensor constructors which can be
        applied to different traces to construct a dataset. For example,
        if you have a `sensor_constructor`:

        ```python
        sensor_constructor = ...
        windowed_sensor_constructor = Window.from_partial_sensor(
            sensor_constructor, ...)

        # ... somewhere inside the dataset constructor
        sensor_instance = windowed_sensor_constructor(path_to_trace)
        ```

        Args:
            sensor: sensor *constructor* to wrap.
            collate_fn: collate function for aggregating a list of samples; if
                not specified, the samples are simply returned as a list.
            past: number of past samples, in addition to the current sample.
                Set to `0` to disable.
            future: number of future samples, in addition to the current
                sample. Set to `0` to disable.
            crop: if `True`, crop the first `past` and last `future` samples in
                the reported metadata to ensure that all samples are full
                valid.
            parallel: maximum number of samples to load in parallel; if `None`,
                all samples are loaded sequentially.
        """
        def create_wrapped_sensor(
            path: str
        ) -> Window[SampleStack, Sample, TMetadata]:
            return cls(
                sensor(path), collate_fn=collate_fn, past=past,
                future=future, crop=crop, parallel=parallel)

        return create_wrapped_sensor

    def __getitem__(self, index: int | np.integer) -> SampleStack:
        """Fetch measurements from this sensor, by index.

        !!! warning

            Note that `past` samples are lost at the beginning, and `future`
            samples at the end to account for the window size!

            If `crop=True`, these lost samples are taken into account by the
            `Window` wrapper; if `crop=False`, the caller must handle this.

        Args:
            index: sample index.

        Returns:
            A set of `past + 1 + future` consecutives samples. Note that there
                is a `past` offset of indices between the wrapped `Window` and
                the underlying sensor!

        Raises:
            IndexError: if `crop=False`, and the requested index is out of
                bounds (i.e., in the first `past` or last `future` samples).
        """
        if self.cropped:
            window = list(range(index, index + self.past + self.future + 1))
        else:
            window = list(range(index - self.past, index + self.future + 1))

        if window[0] < 0 or window[-1] >= len(self.sensor):
            raise IndexError(
                f"Requested invalid index {index} for uncropped "
                f"Window(past={self.past}, future={self.future}).")

        if self.parallel is not None:
            with ThreadPool(min(len(window), self.parallel)) as p:
                return self.collate_fn(p.map(self.sensor.__getitem__, window))
        else:
            return self.collate_fn(list(map(self.sensor.__getitem__, window)))

    def __repr__(self) -> str:
        """Get friendly name (passing through to the underlying sensor)."""
        return f"{repr(self.sensor)} x [-{self.past}:+{self.future}]"


TRaw = TypeVar("TRaw")
TTransformed = TypeVar("TTransformed")
TCollated = TypeVar("TCollated")
TProcessed = TypeVar("TProcessed")


class SequencePipeline(
    spec.Pipeline[
        Sequence[TRaw], Sequence[TTransformed],
        Sequence[TCollated], Sequence[TProcessed]],
    Generic[TRaw, TTransformed, TCollated, TProcessed]
):
    """Transform which passes an additional sequence axis through.

    The given `Pipeline` is modified to accept `Sequence[...]` for each
    data type in its pipeline, and return a `list[...]` across the additional
    axis, thus "passing through" the axis.

    For example, suppose a sequence dataloader reads

    ```
    [
        [Raw[s=0, t=0], Raw[s=0, t=1], ... Raw[s=0, t=n]]
        [Raw[s=1, t=0], Raw[s=1, t=1], ... Raw[s=1, t=n]]
        ...
        [Raw[s=b, t=0], Raw[s=b, t=1], ... Raw[s=b, t=n]
    ]
    ```

    for sequence length `t = 0...n` and batch sample `s = 0...b`. For sequence
    length `t`, the output of the transforms will be batched with the sequence
    on the outside:

    ```
    [
        Processed[s=0...b] [t=0],
        Processed[s=0...b] [t=1],
        ...
        Processed[s=0...b] [t=n]
    ]
    ```

    Type Parameters:
        - `TRaw`, `TTransformed`, `TCollated`, `TProcessed`: see
          [`Pipeline`][abstract_dataloader.spec.].

    Args:
        pipeline: input pipeline.
    """

    def __init__(
        self, pipeline: spec.Pipeline[
            TRaw, TTransformed, TCollated, TProcessed]
    ) -> None:
        self.pipeline = pipeline

    def sample(self, data: Sequence[TRaw]) -> list[TTransformed]:
        return [self.pipeline.sample(x) for x in data]

    def collate(
        self, data: Sequence[Sequence[TTransformed]]
    ) -> list[TCollated]:
        return [self.pipeline.collate(x) for x in zip(*data)]

    def batch(self, data: Sequence[TCollated]) -> list[TProcessed]:
        return [self.pipeline.batch(x) for x in data]



"""ADL specification extensions and non-generic utilities.

Unlike [`abstract_dataloader.generic`][abstract_dataloader.generic], these
implementations "extend" the ADL spec by imposing a particular conceptual
framework on various functionality.

!!! warning

    This module and its submodules are not automatically imported; you will
    need to explicitly import them:

    ```python
    from abstract_dataloader.ext import sample
    ```

!!! info

    This module is not included in the test suite or CI, and is generally held
    to a lower standard than the core `abstract_dataloader`.

The current extension modules are:

- [`augment`][.]: A protocol for specifying data augmentations.
- [`graph`][.]: A programming model based on composing a DAG of callables
    into a single transform.
- [`lightning`][.]: A lightning datamodule wrapper for ADL datasets and
    pipelines.
- [`objective`][.]: Standardized learning objectives, and a
    programming model for multi-objective learning.
- [`sample`][.]: Dataset sampling utilities, including a low-discrepancy
    subset sampler.
- [`types`][.]: Type-related utilities which are not part of the core ADL spec.
"""



"""Tests for ext.graph."""

import pytest
from beartype.claw import beartype_package

beartype_package("abstract_dataloader")

from abstract_dataloader.ext import graph


def test_basic():
    """Test basic graph functionality."""
    def _node_a(a: int, increment: int = 0) -> int:
        return a + 1 + increment

    def _node_b(b: int) -> int:
        return b + 1

    def _node_ab(a: int, b: int) -> int:
        return a + b

    g = {
        "node_a": graph.Node(
            transform=_node_a,
            output="a1",
            inputs={"a": "a_in"},
            optional={"increment": "inc"}
        ),
        "node_b": {
            "transform": _node_b,
            "output": "b1",
            "inputs": {"b": "b_in"},
        },
        "node_ab": graph.Node(
            transform=_node_ab,
            output="ab_out",
            inputs={"a": "a1", "b": "b1"},
        )
    }

    tf_base = graph.Transform(outputs={"out": "ab_out"}, keep_all=False, **g)
    transformed = tf_base({"a_in": 1, "b_in": 2})
    assert transformed == {"out": 5}

    transformed = tf_base({"a_in": 1, "b_in": 2, "inc": 10})
    assert transformed == {"out": 15}

    tf_keepall = graph.Transform(outputs={"out": "ab_out"}, keep_all=True, **g)
    transformed2 = tf_keepall({"a_in": 1, "b_in": 2})
    assert transformed2 == {
        "a_in": 1, "b_in": 2, "a1": 2, "b1": 3, "ab_out": 5, "out": 5}

    tf_no_outputs = graph.Transform(keep_all=False, **g)
    transformed3 = tf_no_outputs({"a_in": 1, "b_in": 2})
    assert transformed3 == {
        "a_in": 1, "b_in": 2, "a1": 2, "b1": 3, "ab_out": 5}


def test_multi_output():
    """Test node with multiple outputs."""
    def _node_multi(a: int) -> tuple[int, int]:
        return a + 1, a + 2

    def _node_single(a: int) -> int:
        return a

    tf = graph.Transform(
        node=graph.Node(
            transform=_node_multi, output=["o1", "o2"], inputs={"a": "a"}))
    transformed = tf({"a": 1})
    assert transformed == {"a": 1, "o1": 2, "o2": 3}

    with pytest.raises(ValueError, match="There are 1 nodes remaining"):
        tf({"x": 0})

    with pytest.raises(ValueError, match="Node 'node' output length mismatch"):
        tf = graph.Transform(
            node=graph.Node(
                transform=_node_multi, output=["o1", "o2", "o3"],
                inputs={"a": "a"}))
        transformed = tf({"a": 1})

    with pytest.raises(TypeError, match="Node 'node' output is expected to be a sequence"):
        tf = graph.Transform(
            node=graph.Node(
                transform=_node_single, output=["o1", "o2"], inputs={"a": "a"}))
        transformed = tf({"a": 1})




"""Abstract Dataloader Generic/Abstract Implementations.

The implementations here provide
[abstract](https://docs.python.org/3/library/abc.html) implementations of
commonly reusable functions such as multi-trace datasets, and glue logic for
synchronization.

- Where applicable, "polyfill" fallbacks also implement some
    methods in terms of more basic ones to allow for extending implementations
    to be more minimal, while still covering required functionality.
- In cases where fallbacks are sufficient to provide a minimal, non-crashing
    implementation of the spec, we omit the [`ABC`][abc.] base class so that
    the class is not technically abstract (though it still may be abstract,
    in the sense that it may not be meaningful to use it directly.)

Some other convenience methods are also provided which are not included in the
core spec; software using the abstract data loader should not rely on these,
and should always base their code on the [`spec`][abstract_dataloader.spec]
types.

!!! fallback

    Abstract base classes which provide default or "fallback" behavior,
    e.g. implementing some methods in terms of others, are documented with
    a `Fallback` section.

!!! note

    Classes without separate abstract implementations are also aliased to
    their original protocol definitions, so that
    [`abstract_dataloader.abstract`][abstract_dataloader.abstract]
    exposes an identical set of objects as
    [`abstract_dataloader.spec`][abstract_dataloader.spec].
"""

from abc import ABC, abstractmethod
from collections.abc import Iterator, Mapping, Sequence
from functools import cached_property
from typing import TypeVar, cast, overload

import numpy as np
from jaxtyping import Int64, Integer

from . import spec
from .spec import Metadata, Synchronization

__all__ = [
    "Dataset", "Metadata", "Sensor", "Synchronization", "Trace", "Pipeline",
    "Transform", "Collate"
]

TSample = TypeVar("TSample")
TMetadata = TypeVar("TMetadata", bound=spec.Metadata)
TTrace = TypeVar("TTrace", bound=spec.Trace)


class Sensor(ABC, spec.Sensor[TSample, TMetadata]):
    """Abstract Sensor Implementation.

    Type Parameters:
        - `TSample`: sample data type which this `Sensor` returns. As a
            convention, we suggest returning "batched" data by default, i.e.
            with a leading singleton axis.
        - `TMetadata`: metadata type associated with this sensor; must
            implement [`Metadata`][abstract_dataloader.spec.].

    Args:
        metadata: sensor metadata, including timestamp information; must
            implement [`Metadata`][abstract_dataloader.spec.].
        name: friendly name; should only be used for debugging and inspection.
    """

    def __init__(self, metadata: TMetadata, name: str = "sensor") -> None:
        self.metadata = metadata
        self.name = name

    @overload
    def stream(self, batch: None = None) -> Iterator[TSample]: ...

    @overload
    def stream(self, batch: int) -> Iterator[list[TSample]]: ...

    def stream(
        self, batch: int | None = None
    ) -> Iterator[TSample | list[TSample]]:
        """Stream values recorded by this sensor.

        Fallback:
            Manually iterate through one sample at a time, loaded using the
            provided `__getitem__` implementation.

        Args:
            batch: batch size; if `0`, returns single samples.

        Returns:
            Iterable of samples (or sequences of samples).
        """
        if batch is None:
            for i in range(len(self)):
                yield self[i]
        else:
            for i in range(len(self) // batch):
                yield [self[j] for j in range(i * batch, (i + 1) * batch)]

    @abstractmethod
    def __getitem__(
        self, index: int | np.integer
    ) -> TSample:
        """Fetch measurements from this sensor, by index.

        Args:
            index: sample index.

        Returns:
            A single sample.
        """
        ...

    def __len__(self) -> int:
        """Total number of measurements.

        Fallback:
            Return the length of the metadata timestamps.
        """
        return self.metadata.timestamps.shape[0]

    @property
    def duration(self) -> float:
        """Trace duration from the first to last sample, in seconds.

        Fallback:
            Compute using the first and last metadata timestamp.
        """
        return self.metadata.timestamps[-1] - self.metadata.timestamps[0]

    @property
    def framerate(self) -> float:
        """Framerate of this sensor, in samples/sec."""
        # `n` samples cover `n-1` periods!
        return (len(self) - 1) / self.duration

    def __repr__(self) -> str:
        """Get friendly representation for inspection and debugging."""
        return f"{self.__class__.__name__}({self.name}, n={len(self)})"


class Trace(spec.Trace[TSample]):
    """A trace, consisting of multiple simultaneously-recording sensors.

    Type Parameters:
        `Sample`: sample data type which this `Sensor` returns. As a
            convention, we suggest returning "batched" data by default, i.e.
            with a leading singleton axis.

    Args:
        sensors: sensors which make up this trace.
        sync: synchronization protocol used to create global samples from
            asynchronous time series. If `Mapping`; the provided indices are
            used directly; if `None`, sensors are expected to already be
            synchronous (equivalent to passing `{k: np.arange(N), ...}`).
        name: friendly name; should only be used for debugging and inspection.
    """

    def __init__(
        self, sensors: Mapping[str, spec.Sensor],
        sync: (
            spec.Synchronization | Mapping[str, Integer[np.ndarray, "N"]]
            | None) = None,
        name: str = "trace"
    ) -> None:
        self.sensors = sensors
        self.name = name

        if sync is None:
            self.indices = None
        elif isinstance(sync, Mapping):
            self.indices = sync
        else:
            self.indices = sync(
                {k: v.metadata.timestamps for k, v in sensors.items()})

    @overload
    def __getitem__(self, index: str) -> Sensor: ...

    @overload
    def __getitem__(self, index: int | np.integer) -> TSample: ...

    def __getitem__(
        self, index: int | np.integer | str
    ) -> TSample | spec.Sensor:
        """Get item from global index (or fetch a sensor by name).

        !!! tip

            For convenience, traces can be indexed by a `str` sensor name,
            returning that [`Sensor`][abstract_dataloader.spec.].

        Fallback:
            Reference implementation which uses the computed
            [`Synchronization`][abstract_dataloader.spec] to retrieve the
            matching indices from each sensor. The returned samples have
            sensor names as keys, and loaded data as values, matching the
            format provided as the `sensors` parameter:

            ```python
            trace[i] = {
                "sensor_a": sensor_a[synchronized_indices["sensor_a"] [i]],
                "sensor_b": sensor_a[synchronized_indices["sensor_b"] [i]],
                ...
            }
            ```

        Args:
            index: sample index, or sensor name.

        Returns:
            Loaded sample if `index` is an integer type, or the appropriate
            [`Sensor`][abstract_dataloader.spec.] if `index` is a `str`.
        """
        if isinstance(index, str):
            return self.sensors[index]

        if self.indices is None:
            return cast(TSample, {
                k: v[index] for k, v in self.sensors.items()})
        else:
            return cast(TSample, {
                k: v[self.indices[k][index].item()]
                for k, v in self.sensors.items()})

    def __len__(self) -> int:
        """Total number of sensor-tuple samples.

        Fallback:
            Returns the number of synchronized index tuples.
        """
        if self.indices is None:
            return len(list(self.sensors.values())[0])
        else:
            return list(self.indices.values())[0].shape[0]

    def __repr__(self) -> str:
        """Friendly representation."""
        sensors = ", ".join(self.sensors.keys())
        return (
            f"{self.__class__.__name__}({self.name}, {len(self)}x[{sensors}])")


class Dataset(spec.Dataset[TSample]):
    """A dataset, consisting of multiple traces, nominally concatenated.

    Type Parameters:
        `Sample`: sample data type which this `Sensor` returns. As a
            convention, we suggest returning "batched" data by default, i.e.
            with a leading singleton axis.

    Args:
        traces: traces which make up this dataset.
    """

    def __init__(self, traces: Sequence[spec.Trace[TSample]]) -> None:
        self.traces = traces

    @cached_property
    def indices(self) -> Int64[np.ndarray, "N"]:
        """End indices of each trace, with respect to global indices."""
        lengths = np.array([len(t) for t in self.traces], dtype=np.int64)
        return np.cumsum(lengths)

    def __getitem__(self, index: int | np.integer) -> TSample:
        """Fetch item from this dataset by global index.

        !!! bug "Unsigned integer subtraction promotes to `np.float64`"

            Subtracting unsigned integers may cause numpy to promote the result
            to a floating point number. Extending implementations should be
            careful about this behavior!

            In the default implementation here, we make sure that the computed
            indices are `int64` instead of `uint64`, and always cast the input
            to an `int64`.

        Fallback:
            Supports (and assumes) random accesses; maps to datasets using
            `np.searchsorted` to search against pre-computed trace start
            indices ([`indices`][^.]), which costs on the order of 10-100us
            per call @ 100k traces.

        Args:
            index: sample index.

        Returns:
            loaded sample.

        Raises:
            IndexError: provided index is out of bounds.
        """
        if index < 0 or index >= len(self):
            raise IndexError(
                f"Index {index} is out of bounds for dataset with length "
                f"{len(self)}.")

        if isinstance(index, np.integer):
            index = np.int64(index)

        trace = np.searchsorted(self.indices, index, side="right")
        if trace > 0:
            remainder = index - self.indices[trace - 1]
        else:
            remainder = index
        # We have to ignore type here since python's Sequence type is not
        # well defined, i.e., does not allow `np.integer` indexing even though
        # `np.integer` is interchangeable with `int`.
        return self.traces[trace][remainder]  # type: ignore

    def __len__(self) -> int:
        """Total number of samples in this dataset.

        Fallback:
            Fetch the dataset length from the trace start indices (at the cost
            of triggering index computation).
        """
        return self.indices[-1].item()

    def __repr__(self) -> str:
        """Friendly representation."""
        return (
            f"{self.__class__.__name__}"
            f"({len(self.traces)} traces, n={len(self)})")


TRaw = TypeVar("TRaw")
TTransformed = TypeVar("TTransformed")
TCollated = TypeVar("TCollated")
TProcessed = TypeVar("TProcessed")


class Transform(spec.Transform[TRaw, TTransformed]):
    """Sample or batch data transform.

    !!! warning

        Transform types are not verified during initialization, and can only
        be verified using runtime type checkers when the transforms are
        applied.

    Type Parameters:
        - `TRaw`: Input data type.
        - `TTransformed`: Output data type.

    Args:
        transforms: transforms to apply sequentially; each output type
            must be the input type of the next transform.
    """

    def __init__(self, transforms: Sequence[spec.Transform]) -> None:
        self.transforms = transforms

    def __call__(self, data: TRaw) -> TTransformed:
        """Apply transforms to a batch of samples.

        Args:
            data: A `TRaw` batch.

        Returns:
            A `TTransformed` batch.
        """
        for tf in self.transforms:
            data = tf(data)
        return cast(TTransformed, data)


class Collate(spec.Collate[TTransformed, TCollated]):
    """Data collation.

    Type Parameters:
        - `TTransformed`: Input data type.
        - `TCollated`: Output data type.
    """

    def __call__(self, data: Sequence[TTransformed]) -> TCollated:
        """Collate a set of samples.

        Args:
            data: A set of `TTransformed` samples.

        Returns:
            A `TCollated` batch.
        """
        return cast(TCollated, data)


class Pipeline(
    spec.Pipeline[TRaw, TTransformed, TCollated, TProcessed]
):
    """Dataloader transform pipeline.

    Composition Rules:
        - A full `Pipeline` can be sequentially pre-composed and/or
          post-composed with one or more [`Transform`][^.]s; this is
          implemented by [`generic.ComposedPipeline`][abstract_dataloader.].
        - `Pipeline`s can always be composed in parallel; this is implemented
          by [`generic.ParallelPipelines`][abstract_dataloader.], with a
          pytorch [`nn.Module`][torch.]-compatible version in
          [`torch.ParallelPipelines`][abstract_dataloader.].

    Type Parameters:
        - `TRaw`: Input data format.
        - `TTransformed`: Data after the first `transform` step.
        - `TCollated`: Data after the second `collate` step.
        - `TProcessed`: Output data format.

    Args:
        sample: sample transform; if `None`, the identity transform is used
            (or the default transform, if overridden).
        collate: sample collation; if `None`, the provided default is used.
            Note that there is no fallback for collation, and
            `NotImplementedError` will be raised if none is provided.
        batch: batch collation; if `None`, the identity transform is used.
    """

    def __init__(
        self, sample: spec.Transform[TRaw, TTransformed] | None = None,
        collate: spec.Collate[TTransformed, TCollated] | None = None,
        batch: spec.Transform[TCollated, TProcessed] | None = None
    ) -> None:
        if sample is not None:
            self.sample = sample
        if collate is not None:
            self.collate = collate
        if batch is not None:
            self.batch = batch

    def sample(self, data: TRaw) -> TTransformed:
        """Transform single samples.

        - Operates on single samples, nominally on the CPU-side of a
          dataloader.
        - This method is both sequentially and parallel composable.

        Fallback:
            The identity transform is provided by default
            (`TTransformed = TRaw`).

        Args:
            data: A single `TRaw` data sample.

        Returns:
            A single `TTransformed` data sample.
        """
        return cast(TTransformed, data)

    def collate(self, data: Sequence[TTransformed]) -> TCollated:
        """Collate a list of data samples into a GPU-ready batch.

        - Operates on the CPU-side of the dataloader, and is responsible for
          aggregating individual samples into a batch (but not transferring to
          the GPU).
        - Analogous to the `collate_fn` of a
          [pytorch dataloader](https://pytorch.org/docs/stable/data.html).
        - This method is not sequentially composable.

        Args:
            data: A sequence of `TTransformed` data samples.

        Returns:
            A `TCollated` collection of the input sequence.
        """
        return cast(TCollated, data)

    def batch(self, data: TCollated) -> TProcessed:
        """Transform data batch.

        - Operates on a batch of data, nominally on the GPU-side of a
          dataloader.
        - This method is both sequentially and parallel composable.

        !!! info "Implementation as `torch.nn.Module`"

            If this `Pipeline` requires GPU state, and the GPU components
            are tied to CPU-side or collation functions (so cannot be
            separated and implemented separately) it may be helpful to
            implement the `Pipeline` as a `torch.nn.Module`. In this case,
            `batch` should redirect to `__call__`, which in turn redirects to
            [`nn.Module.forward`][torch.] in order to handle any registered
            pytorch hooks.

        Fallback:
            The identity transform is provided by default
            (`TProcessed = TCollated`).

        Args:
            data: A `TCollated` batch of data, nominally already sent to the
                GPU.

        Returns:
            The `TProcessed` output, ready for the downstream model.
        """
        return cast(TProcessed, data)



"""Transform compositions."""

from collections.abc import Sequence
from typing import Any, Generic, TypeVar, cast

from abstract_dataloader import spec

PRaw = TypeVar("PRaw", bound=dict[str, Any])
PTransformed = TypeVar("PTransformed", bound=dict[str, Any])
PCollated = TypeVar("PCollated", bound=dict[str, Any])
PProcessed = TypeVar("PProcessed", bound=dict[str, Any])


class ParallelTransforms(spec.Transform[PRaw, PTransformed]):
    """Compose multiple transforms, similar to [`ParallelPipelines`][^.].

    Type Parameters:
        - `PRaw`, `PTransformed`, [`Transform`][abstract_dataloader.spec.].

    Args:
        transforms: transforms to compose. The key indicates the subkey to
            apply each transform to.
    """

    def __init__(self, **transforms: spec.Transform) -> None:
        self.transforms = transforms

    def __call__(self, data: PRaw) -> PTransformed:
        return cast(
            PTransformed,
            {k: v(data[k]) for k, v in self.transforms.items()})


class ParallelPipelines(
    spec.Pipeline[PRaw, PTransformed, PCollated, PProcessed],
):
    """Compose multiple transforms in parallel.

    For example, with transforms `{"radar": radar_tf, "lidar": lidar_tf, ...}`,
    the composed transform performs:

    ```python
    {
        "radar": radar_tf.transform(data["radar"]),
        "lidar": lidar_tf.transform(data["lidar"]),
        ...
    }
    ```

    !!! note

        This implies that the type parameters must be `dict[str, Any]`, so this
        class is parameterized by a separate set of
        `Composed(Raw|Transformed|Collated|Processed)` types with this bound.

    !!! tip

        See [`torch.ParallelPipelines`][abstract_dataloader.] for an
        implementation which is compatible with [`nn.Module`][torch.]-based
        pipelines.

    Type Parameters:
        - `PRaw`, `PTransformed`, `PCollated`, `PProcessed`: see
          [`Pipeline`][abstract_dataloader.spec.].

    Args:
        transforms: transforms to compose. The key indicates the subkey to
            apply each transform to.
    """

    def __init__(self, **transforms: spec.Pipeline) -> None:
        self.transforms = transforms

    def sample(self, data: PRaw) -> PTransformed:
        return cast(
            PTransformed,
            {k: v.sample(data[k]) for k, v in self.transforms.items()})

    def collate(self, data: Sequence[PTransformed]) -> PCollated:
        return cast(PCollated, {
            k: v.collate([x[k] for x in data])
            for k, v in self.transforms.items()
        })

    def batch(self, data: PCollated) -> PProcessed:
        return cast(
            PProcessed,
            {k: v.batch(data[k]) for k, v in self.transforms.items()})


TRawInner = TypeVar("TRawInner")
TRaw = TypeVar("TRaw")
TTransformed = TypeVar("TTransformed")
TCollated = TypeVar("TCollated")
TProcessed = TypeVar("TProcessed")
TProcessedInner = TypeVar("TProcessedInner")


class ComposedPipeline(
    spec.Pipeline[TRaw, TTransformed, TCollated, TProcessed],
    Generic[
        TRaw, TRawInner, TTransformed,
        TCollated, TProcessedInner, TProcessed]
):
    """Compose pipeline sequentially with pre and post transforms.

    Type Parameters:
        - `TRaw`: initial input type.
        - `TRawInner`: output of the pre-composed transform, and input to the
            provided [`Pipeline`][abstract_dataloader.spec].
        - `TCollated`, `TProcessed`: intermediate values for the provided
            [`Pipeline`][abstract_dataloader.spec].
        - `TProcessedInner`: output of the transforms, and input to the
            post-composed transform.
        - `TProcessed`: output type.

    Args:
        pipeline: pipeline to compose.
        pre: pre-transform to apply on the CPU side; skipped if `None`.
        post: post-transform to apply on the GPU side; skipped if `None`.
    """

    def __init__(
        self, pipeline: spec.Pipeline[
            TRawInner, TTransformed, TCollated, TProcessedInner],
        pre: spec.Transform[TRaw, TRawInner] | None = None,
        post: spec.Transform[TProcessedInner, TProcessed] | None = None
    ) -> None:
        self.pipeline = pipeline
        self.pre = pre
        self.post = post

        self.collate = pipeline.collate

    def sample(self, data: TRaw) -> TTransformed:
        """Transform single samples.

        Args:
            data: A single `TRaw` data sample.

        Returns:
            A single `TTransformed` data sample.
        """
        if self.pre is None:
            transformed = cast(TRawInner, data)
        else:
            transformed = self.pre(data)
        return self.pipeline.sample(transformed)

    def batch(self, data: TCollated) -> TProcessed:
        """Transform data batch.

        Args:
            data: A `TCollated` batch of data, nominally already sent to the
                GPU.

        Returns:
            The `TProcessed` output, ready for the downstream model.
        """
        transformed = self.pipeline.batch(data)
        if self.post is None:
            return cast(TProcessed, transformed)
        else:
            return self.post(transformed)



"""Dataloader / Pytorch Bridge.

!!! warning

    [Pytorch lightning](https://lightning.ai/docs/pytorch/stable/) must be
    installed to use this module. This is not included in any extras; you will
    need to `pip install lightning` or add it to your dependencies.

The provided data module is based on the following assumptions:

- All splits use the same transform [`Pipeline`][abstract_dataloader.spec.],
    but each have a different [`Dataset`][abstract_dataloader.spec.Dataset].
    This means that if any data augmentations are applied by the transform, the
    `Dataset` should pass some `meta` information (i.e., whether in training
    mode) as part of the data.
- In-training visualizations are always rendered from the same set of a
    relatively small number of samples taken from the validation set.
- The same dataloader settings should be applied to all splits.

!!! info

    Only sample-to-sample (`.transform`) and sample-to-batch (`.collate`)
    transforms are applied in the dataloader; the training loop is
    responsible for applying batch-to-batch (`.forward`) transforms.
"""

from functools import cache, cached_property, partial
from typing import Any, Callable, Generic, Literal, Mapping, Sequence, TypeVar

import lightning as L  # noqa: N812
import numpy as np
from torch.utils.data import DataLoader

from abstract_dataloader import spec
from abstract_dataloader import torch as adl_torch

from .sample import SampledDataset

Sample = TypeVar("Sample")
Raw = TypeVar("Raw", contravariant=True, bound=dict[str, Any])
Transformed = TypeVar("Transformed", bound=dict[str, Any])
Collated = TypeVar("Collated", bound=dict[str, Any])
Processed = TypeVar("Processed", covariant=True, bound=dict[str, Any])


class ADLDataModule(
    L.LightningDataModule, Generic[Raw, Transformed, Collated, Processed]
):
    """Pytorch dataloader wrapper for ADL-compliant datasets.

    !!! info

        Train/val/test splits are not all required to be present; if any are
        not present, the corresponding `.{split}_dataloader()` will raise an
        error if called. Arbitrary split names are also allowed, though
        `train`, `val`, and `test` are expected for the
        `ADLDataModule.{train|val|test}_dataloader()` methods
        [expected by pytorch lightning](
        https://lightning.ai/docs/pytorch/stable/data/datamodule.html).

    !!! note

        The underlying (transformed) dataset is cached (i.e. the same
        dataset object will be used on each call), but the dataloader
        container is not.

    Type Parameters:
        - `Raw`: raw data loaded from the dataset.
        - `Transformed`: data following CPU-side transform.
        - `Collated`: data format after collation; should be in pytorch tensors.
        - `Processed`: data after GPU-side transform.

    Args:
        dataset: datasets or dataset constructors for each split.
        transforms: data transforms to apply.
        batch_size: dataloader batch size.
        samples: number of validation-set samples to prefetch for
            visualizations (or a list of indices to use). Note that these
            samples are always held in memory! Set `samples=0` to disable.
        num_workers: number of worker processes during data loading and
            CPU-side processing; use `num_workers=0` to run in the main thread.
        prefetch_factor: number of batches to fetch per worker. Must be `None`
            when `num_workers=0`.
        subsample: Sample only a (low-discrepancy) subset of samples on each
            split specified here instead of using all samples.

    Attributes:
        transforms: data transforms which should be applied to the data; in
            particular, a `.forward()` GPU batch-to-batch stage which is
            expected to be handled by downstream model code.
    """

    def __init__(
        self, dataset: Mapping[
            str, Callable[[], spec.Dataset[Raw]] | spec.Dataset[Raw]],
        transforms: spec.Pipeline[Raw, Transformed, Collated, Processed],
        batch_size: int = 32, samples: int | Sequence[int] = 0,
        num_workers: int = 32, prefetch_factor: int | None = None,
        subsample: Mapping[str, int | float | None] = {}
    ) -> None:
        super().__init__()

        self.transforms = transforms
        self._samples = samples
        self._subsample = subsample

        self._dataset = dataset
        self._dataloader_args = {
            "batch_size": batch_size, "num_workers": num_workers,
            "prefetch_factor": prefetch_factor, "pin_memory": True,
            "collate_fn": transforms.collate
        }

    @classmethod
    def from_traces(
        cls, dataset: Callable[[Sequence[str]], spec.Dataset[Raw]],
        traces: Mapping[str, Sequence[str]],
        transforms: spec.Pipeline[Raw, Transformed, Collated, Processed],
        **kwargs: dict[str, Any]
    ) -> "ADLDataModule[Raw, Transformed, Collated, Processed]":
        """Create from a dataset constructor.

        Args:
            dataset: dataset constructor which takes a list of trace names
                and returns a dataset object.
            traces: mapping of split names to trace names; the dataset
                constructor will be called with the trace names for each split.
            transforms: data transforms to apply.
            kwargs: see the class constructor.
        """
        return cls(
            dataset={k: partial(dataset, v) for k, v in traces.items()},
            transforms=transforms,
            **kwargs)  # type: ignore

    @cached_property
    def samples(self) -> Collated | None:
        """Validation samples for rendering samples.

        If a simple `samples: int` is specified, these samples are taken
        uniformly `len(val) // samples` apart with padding on either side.

        !!! warning

            While this property is cached, accessing this property the first
            time triggers a full load of the dataset validation split!

        Returns:
            Pre-loaded validation samples, nominally for generating
                visualizations. If `samples=0` was specified, or no validation
                split is provided, then no samples are returned.
        """
        if self._samples != 0:
            try:
                val = self.dataset("val")

                if isinstance(self._samples, int):
                    m = len(val) // self._samples // 2
                    indices = np.linspace(
                        m, len(val) - m, self._samples, dtype=np.int64)
                else:
                    indices = self._samples

                sampled = [val[i] for i in indices]
                return self.transforms.collate(sampled)
            except KeyError:
                return None
        return None

    @cache
    def dataset(
        self, split: Literal["train", "val", "test"] = "train"
    ) -> adl_torch.TransformedDataset[Raw, Transformed]:
        """Get dataset for a given split, with sample transformation applied.

        !!! info

            If the a split is requested, and `subsample` is specified for that
            split, a subsample transform (via
            [`SampledDataset`][abstract_dataloader.ext.sample.]) is also
            applied.

        Args:
            split: target split.

        Returns:
            Dataset for that split, using the partially bound constructor
                passed to the `ADLDataModule`; the dataset is cached between
                calls.
        """
        if split not in self._dataset:
            raise KeyError(
                f"No `{split}` split was provided to this DataModule. Only "
                f"the following splits are present: "
                f"{list(self._dataset.keys())}")

        dataset = self._dataset[split]
        if not isinstance(dataset, spec.Dataset):
            dataset = dataset()

        subsample = self._subsample.get(split)
        if subsample is not None:
            dataset = SampledDataset(dataset, subsample)

        return adl_torch.TransformedDataset(
            dataset, transform=self.transforms.sample)

    def train_dataloader(self) -> DataLoader:
        """Get training dataloader (`shuffle=True, drop_last=True`)."""
        return DataLoader(
            self.dataset("train"), shuffle=True, drop_last=True,
            **self._dataloader_args)

    def val_dataloader(self) -> DataLoader:
        """Get validation dataloader (`shuffle=False, drop_last=True`)."""
        return DataLoader(
            self.dataset("val"), shuffle=False, drop_last=True,
            **self._dataloader_args)

    def test_dataloader(self) -> DataLoader:
        """Get test dataloader (`shuffle=False, drop_last=False`)."""
        return DataLoader(
            self.dataset("test"), shuffle=False, drop_last=False,
            **self._dataloader_args)



"""Pytorch module registration."""

import numpy as np
import torch
from beartype.claw import beartype_package
from torch import nn

beartype_package("abstract_dataloader")

from abstract_dataloader import abstract  # noqa: E402
from abstract_dataloader import torch as adl_torch  # noqa: E402


def test_composition():
    """Test that module composition is correctly applied across pipelines."""

    class _BatchModule(nn.Module):

        def __init__(self) -> None:
            super().__init__()
            self.parameter = nn.Parameter(torch.zeros(()))

    class _PipelineModule(nn.Module):

        def __init__(self) -> None:
            super().__init__()
            self.collate = adl_torch.Collate(mode='stack')
            self._batch = _BatchModule()

        def forward(self, data):
            raise NotImplementedError

        def sample(self, data):
            raise NotImplementedError

        def batch(self, data):
            raise NotImplementedError

    p1 = _PipelineModule()
    p2 = _PipelineModule()

    assert set(p1.parameters()) == set(p1._batch.parameters())

    sequenced = adl_torch.SequencePipeline(p1)
    assert isinstance(sequenced, nn.Module)
    assert set(sequenced.parameters()) == set(p1.parameters())

    paralleled = adl_torch.ParallelPipelines(p1=p1, p2=p2)
    assert set(paralleled.parameters()) == set(p1.parameters()).union(
        p2.parameters())

    composed = adl_torch.ComposedPipeline(
        pipeline=p1, pre=lambda x: x, post=lambda x: x)  # type: ignore
    assert set(composed.parameters()) == set(p1.parameters())


def test_collate():
    """Test collate function."""
    data = {
        "a": np.random.default_rng().uniform(size=(3, 5)),
        "b": np.random.default_rng().uniform(size=(3, 4))
    }

    collate1 = adl_torch.Collate(mode='stack')
    c1 = collate1([{"a": data["a"][0], "b": data["b"][0]},
              {"a": data["a"][1], "b": data["b"][1]},
              {"a": data["a"][2], "b": data["b"][2]}])
    assert isinstance(c1["a"], torch.Tensor)
    assert isinstance(c1["b"], torch.Tensor)
    assert torch.all(c1["a"] == torch.from_numpy(data["a"]))
    assert torch.all(c1["b"] == torch.from_numpy(data["b"]))

    collate2 = adl_torch.Collate(mode='concat')
    c2 = collate2([{"a": data["a"][0][None], "b": data["b"][0][None]},
              {"a": data["a"][1][None], "b": data["b"][1][None]},
              {"a": data["a"][2][None], "b": data["b"][2][None]}])
    assert isinstance(c2["a"], torch.Tensor)
    assert isinstance(c2["b"], torch.Tensor)
    assert torch.all(c2["a"] == torch.from_numpy(data["a"]))
    assert torch.all(c2["b"] == torch.from_numpy(data["b"]))


def test_stacked_sequence():
    """Test StackedSequencePipeline transform."""
    pipeline = abstract.Pipeline(
        sample=lambda data: {
            "a": np.array(data["a"]),  # type: ignore
            "b": np.array(data["b"])   # type: ignore
        },
        collate=lambda data: {
            "a": torch.stack([torch.tensor(d["a"]) for d in data]),
            "b": torch.stack([torch.tensor(d["b"]) for d in data]),
        },
        batch=lambda data: {
            "a": data["a"] * 2,
            "b": data["b"] * 3,
        },
    )
    stacked_pipeline = adl_torch.StackedSequencePipeline(pipeline)

    raw = [
        [{"a": [1, 2], "b": [3, 4]}, {"a": [5, 6], "b": [7, 8]}],
        [{"a": [9, 10], "b": [11, 12]}, {"a": [13, 14], "b": [15, 16]}],
    ]

    transformed = [stacked_pipeline.sample(x) for x in raw]
    collated = stacked_pipeline.collate(transformed)
    processed = stacked_pipeline.batch(collated)

    assert isinstance(processed["a"], torch.Tensor)
    assert isinstance(processed["b"], torch.Tensor)
    assert processed["a"].shape == (2, 2, 2)  # batch, sequence, features
    assert processed["b"].shape == (2, 2, 2)
    assert torch.all(processed["a"] == torch.tensor(
        [[[2, 4], [10, 12]], [[18, 20], [26, 28]]]))
    assert torch.all(processed["b"] == torch.tensor(
        [[[9, 12], [21, 24]], [[33, 36], [45, 48]]]))


def test_transformed_dataset():
    """Test TransformedDataset functionality."""
    class MockDataset:
        def __init__(self, data):
            self.data = data

        def __getitem__(self, index):
            return self.data[index]

        def __len__(self):
            return len(self.data)

    def mock_transform(sample):
        return sample * 2

    raw_data = [1, 2, 3]
    dataset = MockDataset(raw_data)
    transformed_dataset = adl_torch.TransformedDataset(
        dataset, mock_transform)  # type: ignore

    assert len(transformed_dataset) == len(raw_data)
    for i in range(len(transformed_dataset)):
        assert transformed_dataset[i] == raw_data[i] * 2



"""Dataset sampling, including a low discrepancy subset sampler.

Dataset sampling is implemented using a [`SampledDataset`][.],
which transparently wraps an existing [`Dataset`][abstract_dataloader.spec.].
"""

from typing import Callable, Generic, Literal, TypeVar

import numpy as np
from jaxtyping import Int64, Integer

from abstract_dataloader import spec

TSample = TypeVar("TSample")


class SampledDataset(spec.Dataset[TSample], Generic[TSample]):
    """Dataset wrapper which only exposes a subset of values.

    The sampling `mode` can be one of:

    - `random`: Uniform random sampling, with `np.random.default_rng` and the
      supplied seed; if `seed` is a `float`, it is converted into an integer
      by multiplying by `len(dataset)` and rounding.
    - `ld`: Low discrepancy sampling; see [`sample_ld`][^.].
    - `uniform`: Uniformly spaced sampling, with `linspace(0, n, samples)`.
    - `Callable`: A callable which takes the total number of samples, and
        returns an array of indices to sample from the dataset.

    !!! info

        This `SampledDataset` is fully ADL-compliant, and acts as a passthrough
        to an ADL-compliant [`Dataset`][abstract_dataloader.spec.]: if the
        input dataset is a `Dataset[Sample]`, then the wrapped dataset is also
        a `Dataset[Sample]`.

    Type Parameters:
        `Sample`: dataset sample type.

    Args:
        dataset: underlying dataset.
        samples: target number of samples.
        seed: sampler seed.
        mode: sampling mode.
    """

    def __init__(
        self, dataset: spec.Dataset[TSample], samples: int | float,
        seed: int | float = 0,
        mode: Literal["ld", "uniform", "random"]
            | Callable[[int], Integer[np.ndarray, "N"]] = "ld"
    ) -> None:
        self.dataset = dataset

        if isinstance(samples, float):
            samples = int(samples * len(dataset))

        if mode == "ld":
            self.subset = sample_ld(len(dataset), samples=samples, seed=seed)
        elif mode == "random":
            if isinstance(seed, float):
                seed = int(seed * len(dataset))
            self.subset = np.random.default_rng(seed).choice(
                len(dataset), size=samples, replace=True)
        elif mode == "uniform":
            self.subset = np.linspace(
                0, len(dataset) - 1, samples, dtype=np.int64)
        else:  # Callable
            self.subset = mode(len(dataset)).astype(np.int64)

    def __getitem__(self, index: int | np.integer) -> TSample:
        """Fetch item from this dataset by global index."""
        return self.dataset[self.subset[index]]

    def __len__(self) -> int:
        """Total number of samples in this dataset."""
        return self.subset.shape[0]


def sample_ld(
    total: int, samples: float | int,
    seed: float | int = 0, alpha: float | int = 2 / (np.sqrt(5) + 1),
) -> Int64[np.ndarray, "samples"]:
    """Compute deterministic low-discrepancy subset mask.

    Uses a simple `alpha * n % 1` formulation, described
    [here](https://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/),
    with a modification to work with integer samples:

    - For a given `total`, find the integer closest to `total * alpha` which
      is co-prime with the total. Use this as the step size.
    - Then, `1...total * alpha (mod total)` is guaranteed to visit each index
      up to `total` exactly once.

    !!! note

        The default `alpha = 1 / phi` where `phi` is the golden ratio
        `(1 + sqrt(5)) / 2` has strong [low-discrepancy sampling properties
        ](https://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/);
        due to the quantized nature of this function, the discrepancy may be
        larger when `total` is small.

    !!! tip

        Each of the parameters (`samples`, `seed`, `alpha`) can be
        specified as a float `[0, 1]`, and a proportion of the `total` will
        be used instead. For example, if `seed = 0.7` and `total=100`, then
        `seed = 70` will be used.

    Args:
        total: total number of samples to sample from, i.e. maximum index.
        samples: number of samples to generate. Should be less than `total`.
        seed: initial offset for the sampling sequence. Can leave this at `0`.
        alpha: step size in the sequence; the default value is the inverse
            golden ratio `2 / (np.sqrt(5) + 1)` (which is actually equivalent
            to the golden ratio `mod 1`, since `1 - phi = 1 / phi`).

    Returns:
        Array, in mixed order, of `sample` indices which form a subset of
            `0...total` with a guarantee of no repeats.
    """
    def _get_qld_step(n: int, q: int) -> int:
        """Get quantized low-discrepancy step size."""
        for i in range(min(n - q, q)):
            query =  q + (((i + 1) // 2) * (-1 if i % 2 == 0 else 1))
            if np.gcd(query, n) == 1:
                return query
        raise ValueError(
            "No GCD found. This should be impossible! Is `n` degenerate?")

    if isinstance(samples, float):
        samples = int(samples * total)
    if isinstance(seed, float):
        seed = int(seed * total)
    if isinstance(alpha, float):
        alpha = _get_qld_step(total, int(total * alpha))

    if samples > total or samples < 0:
        raise ValueError(
            f"Number of samples {samples} must be in [0, {total}].")
    if samples * alpha >= np.iinfo(np.int64).max:
        raise NotImplementedError(
            f"`samples={samples}` is too large, and will cause overflow.")

    return (np.arange(samples, dtype=np.int64) * alpha + seed) % total



"""Generic "ready-to-use" implementations of common components.

Other generic and largely reusable components can be added to this submodule.

!!! note

    Numpy (and jaxtyping) are the only dependencies; to keep the
    `abstract_dataloader`'s dependencies lightweight and flexible, components
    should only be added here if they do not require any additional
    dependencies.
"""

from .composition import ComposedPipeline, ParallelPipelines, ParallelTransforms
from .sequence import Metadata, SequencePipeline, Window
from .sync import Empty, Nearest, Next

__all__ = [
    "ComposedPipeline", "ParallelPipelines", "ParallelTransforms",
    "Metadata", "SequencePipeline", "Window",
    "Empty", "Nearest", "Next"
]



"""Composing transforms for data processing pipelines.

!!! abstract "Programming Model"

    - Data is represented as a dictionary with string keys and arbitrary
        values which are atomic from the perspective of transform composition.
    - Transforms are created from a directed acyclic graph (DAG) of nodes,
        where each node ([`Node`][.]) is a callable which takes a set
        of inputs and produces a set of outputs.
"""

from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass, field
from typing import Any

import wadler_lindig as wl

from abstract_dataloader import spec


@dataclass
class Node:
    """Node specification for a graph-based data processing transform.

    !!! example "Example Hydra Config"

        ```yaml
        rsp:
          inputs:
            iq: radar
            aug: aug
          output: spectrum
          transform:
            _target_: grt.transforms.Spectrum
            ...
        ```

    Attributes:
        transform: callable to apply to the inputs.
        output: output data key (or output data keys for a node which returns
            multiple outputs).
        inputs: mapping of data keys to input argument names.
        optional: mapping of optional data keys to input argument names (i.e.,
            they are only passed if present).
    """

    transform: Callable
    output: str | Sequence[str]
    inputs: Mapping[str, str] = field(default_factory=dict)
    optional: Mapping[str, str] = field(default_factory=dict)

    def apply(self, data: dict[str, Any], name: str = "") -> dict[str, Any]:
        """Apply the node.

        Args:
            data: input data to process.
            name: node name (for error messages).

        Returns:
            Updated data, with any new keys added to the input data.
        """
        inputs = {k: data[v] for k, v in self.inputs.items()}
        for k, v in self.optional.items():
            if v in data:
                inputs[k] = data[v]

        output = self.transform(**inputs)

        if isinstance(self.output, str):
            data[self.output] = output
        else:  # Sequence[str]
            if not isinstance(output, Sequence):
                raise TypeError(
                    f"Node '{name}' output is expected to be a sequence due "
                    f"to output specification {self.output}: "
                    f"\n{wl.pformat(output)}\n")
            if len(self.output) != len(output):
                raise ValueError(
                    f"Node '{name}' output length mismatch: expected "
                    f"{len(self.output)} outputs ({self.output}), but got "
                    f"{len(output)} outputs:\n{wl.pformat(output)}\n")

            for o, v in zip(self.output, output):
                data[o] = v

        return data


class Transform(spec.Transform[dict[str, Any], dict[str, Any]]):
    """Compose multiple callables forming a DAG into a transform.

    !!! warning

        Since the input data specifications are not provided at initialization,
        the graph execution order (or graph validity) is not statically
        determined, and result in runtime errors if invalid.

    Args:
        outputs: output data keys to produce as a mapping of output keys to
            graph data keys. If `None`, all values are returned.
        keep_all: keep references to all intermediate values and return them
            instead of decref-ing values which are no longer needed.
        nodes: nodes in the graph, as keyword arguments where the key indicates
            a reference name for the node; any `dict` arguments are passed as
            key/value arguments to [`Node`][^.].
    """

    def __init__(
        self, outputs: Mapping[str, str] | None = None, keep_all: bool = False,
        **nodes: Node | dict[str, Any]
    ) -> None:
        self.nodes = {
            k: v if isinstance(v, Node) else Node(**v)
            for k, v in nodes.items()}
        self.outputs = outputs
        self.keep_all = keep_all

    def _err_disconnected(
        self, data: dict[str, Any], incomplete: dict[str, Node]
    ) -> ValueError:
        """Format error message for disconnected nodes."""
        remaining = {k: list(v.inputs.values()) for k, v in incomplete.items()}
        return ValueError(
            f"There are {len(incomplete)} nodes remaining, but "
            f"all remaining nodes have at least one missing input.\n"
            f"Current inputs: {list(data.keys())}\n"
            f"Remaining stage requirements: {wl.pformat(remaining)}")

    def _decref(
        self, data: dict[str, Any], incomplete: dict[str, Node]
    ) -> dict[str, Any]:
        """Decref unneeded data values."""
        if self.keep_all or self.outputs is None:
            return data
        else:
            keep = set(self.outputs.values())
            for node in self.nodes.values():
                keep |= set(node.inputs.values())
            return {k: v for k, v in data.items() if k in keep}

    def __call__(self, data: dict[str, Any]) -> dict[str, Any]:
        """Execute the transform graph on the input data.

        Args:
            data: input data to process.

        Returns:
            Processed data.
        """
        incomplete = self.nodes.copy()

        # Guaranteed to terminate:
        # Each loop removes one node, or raises an error.
        while len(incomplete) > 0:
            for name, node in incomplete.items():
                if all(k in data for k in node.inputs.values()):
                    data = node.apply(data, name=name)
                    incomplete.pop(name)
                    data = self._decref(data, incomplete)
                    break  # break back into the while
            else:
                raise self._err_disconnected(data, incomplete)

        if self.outputs is not None:
            if self.keep_all:
                for k, v in self.outputs.items():
                    data[k] = data[v]
                return data
            else:
                return {k: data[v] for k, v in self.outputs.items()}
        else:
            return data



"""Windowing and sequence transforms."""

import numpy as np
import pytest
from beartype.claw import beartype_package

beartype_package("abstract_dataloader")

from abstract_dataloader import abstract, generic, spec  # noqa: E402


class _ConcreteMetadata(spec.Metadata):

    def __init__(self, n: int = 10):
        self.timestamps = np.arange(n, dtype=np.float64)

class _ConcreteSensor(abstract.Sensor[int, _ConcreteMetadata]):

    def __getitem__(self, index: int | np.integer) -> int:
        return int(index)


def _apply_pipeline(pipeline, raw):
    transformed = [pipeline.sample(x) for x in raw]
    collated = pipeline.collate(transformed)
    processed = pipeline.batch(collated)
    return processed


@pytest.mark.parametrize("parallel", [None, 1, 2])
def test_window(parallel):
    """Test windowing."""
    meta = _ConcreteMetadata(5)

    partialed = generic.Window.from_partial_sensor(
        lambda path: _ConcreteSensor(meta),
        past=1, future=1, parallel=parallel)
    windowed = partialed("ignored")

    assert windowed[0] == [0, 1, 2]
    assert windowed[1] == [1, 2, 3]
    assert windowed[2] == [2, 3, 4]
    assert np.all(windowed.metadata.timestamps == np.array([1, 2, 3]))


@pytest.mark.parametrize("parallel", [None, 1, 2])
def test_window_uncropped(parallel) -> None:
    """Test uncropped windowing."""
    meta = _ConcreteMetadata(5)

    partialed = generic.Window.from_partial_sensor(
        lambda path: _ConcreteSensor(meta),
        past=1, future=1, crop=False, parallel=parallel)
    windowed = partialed("ignored")

    assert windowed[1] == [0, 1, 2]
    assert windowed[2] == [1, 2, 3]
    assert windowed[3] == [2, 3, 4]
    assert np.all(windowed.metadata.timestamps == np.array([0, 1, 2, 3, 4]))

    with pytest.raises(IndexError):
        _ = windowed[0]
    with pytest.raises(IndexError):
        _ = windowed[4]


def test_transform():
    """Test sequence transforms."""
    pipeline = abstract.Pipeline(
        sample=lambda data: data * 2,  # type: ignore
        collate=lambda data: data,
        batch=lambda data: [x * 5 for x in data])
    sequenced = generic.SequencePipeline(pipeline)

    raw = [[1, 4], [2, 5], [3, 6]]
    processed = _apply_pipeline(sequenced, raw)
    assert processed == [[10, 20, 30], [40, 50, 60]]


def test_parallel():
    """Test parallel transforms."""
    pipeline_a = abstract.Pipeline(
        sample=lambda data: data * 2,  # type: ignore
        collate=lambda data: data,
        batch=lambda data: [x * 5 for x in data])

    pipeline_b = abstract.Pipeline(
        sample=lambda data: data + 1,  # type: ignore
        collate=lambda data: data,
        batch=lambda data: [x + 2 for x in data])

    composed = generic.ParallelPipelines(a=pipeline_a, b=pipeline_b)
    raw = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
    processed = _apply_pipeline(composed, raw)
    assert processed == {"a": [10, 30], "b": [5, 7]}


def test_sequential():
    """Test sequential composition."""
    pipeline = abstract.Pipeline(collate=lambda data: data)
    pre = lambda data: data * 2
    post = lambda data: [x * 5 for x in data]
    raw = [1, 2, 3]

    composed = generic.ComposedPipeline(pipeline=pipeline, pre=pre, post=post)
    processed = _apply_pipeline(composed, raw)
    assert processed == [10, 20, 30]

    composed = generic.ComposedPipeline(pipeline=pipeline, pre=pre)
    processed = _apply_pipeline(composed, raw)
    assert processed == [2, 4, 6]

    composed = generic.ComposedPipeline(pipeline=pipeline, post=post)
    processed = _apply_pipeline(composed, raw)
    assert processed == [5, 10, 15]

    composed = generic.ComposedPipeline(pipeline=pipeline)
    assert _apply_pipeline(composed, raw) == _apply_pipeline(pipeline, raw)



"""Abstract specifications for data augmentations.

!!! abstract "Programming Model"

    - Data augmentations consist of a set of properties which correspond to
        common properties of the data, such as `azimuth_flip` or `range_scale`.
    - The user is responsible for defining and keeping track of each
        augmentation property and its meaning.
    - Multiple data augmentation specifications can be combined into a single
        set of `Augmentations`, which generates a dictionary of values.

In addition to the general framework, we include wrappers for a few common
distributions:

- [`Bernoulli`][.]: `Bernoulli(p)`
- [`Normal`][.]: `Normal(0, std) * Bernoulli(p)`
- [`TruncatedLogNormal`][.]:
    `exp(clamp(Normal(0, std), -clip, clip)) * Bernoulli(p))`
- [`Uniform`][.]: `Unif(lower, upper) * Bernoulli(p)`
"""

import os
from typing import Any, Generic, Protocol, TypeVar

import numpy as np

T = TypeVar("T", covariant=True)

class Augmentation(Protocol, Generic[T]):
    """A generic augmentation random generation policy."""

    def __call__(self, rng: np.random.Generator) -> T:
        """Sample the value of a data augmentation parameter."""
        ...


class Augmentations:
    """A collection of data augmentations.

    Args:
        kwargs: augmentations to apply, where each key is the
            name of a physical property (e.g., `azimuth_flip`, `range_scale`)
            and each value is the corresponding augmentation generator.
    """

    def __init__(self, **kwargs: Augmentation) -> None:
        self.augmentations = kwargs
        self._rng = {}

    @property
    def rng(self) -> np.random.Generator:
        """Random number generator, using the PID as a seed."""
        pid = os.getpid()
        if pid in self._rng:
            return self._rng[pid]
        else:
            rng = np.random.default_rng(seed=pid)
            self._rng[pid] = rng
            return rng

    def __call__(self, meta: dict[str, Any] = {}) -> dict[str, Any]:
        """Generate a dictionary of augmentations.

        If a `train=False` flag is passed in `meta`, no augmentations are
        generated.

        Args:
            meta: data processing configuration inputs.

        Returns:
            Augmentation specifications.
        """
        if meta.get("train", True):
            return {k: v(self.rng) for k, v in self.augmentations.items()}
        else:
            return {}


class Bernoulli(Augmentation[bool]):
    """Enable augmentation with certain probability.

    Type: `bool` (`True` if enabled).

    Args:
        p: probability of enabling.
    """

    def __init__(self, p: float = 0.5) -> None:
        self.p = p

    def __call__(self, rng: np.random.Generator) -> bool:
        return rng.random() < self.p


class Normal(Augmentation[float]):
    """Normal distribution.

    Type: `float`; returns `0.0` if not enabled. Always zero-centered.

    Args:
        p: probability of enabling this augmentation (`True`).
        std: standard deviation of the normal distribution.
    """

    def __init__(self, p: float = 1.0, std: float = 1.0) -> None:
        self.p = p
        self.std = std

    def __call__(self, rng: np.random.Generator) -> float:
        if self.p < 1.0 and rng.random() > self.p:
            return 0.0

        return rng.normal(scale=self.std)


class TruncatedLogNormal(Augmentation[float]):
    """Truncated log-normal distribution.

    The underlying normal is always centered around zero.

    Type: `float`; returns `1.0` if not enabled.

    Args:
        p: probability of enabling this augmentation (`True`).
        std: standard deviation of the underlying normal distribution.
        clip: clip to this many standard deviations; don't clip if 0.
    """

    def __init__(
        self, p: float = 1.0, std: float = 0.2, clip: float = 2.0
    ) -> None:
        self.p = p
        self.std = std
        self.clip = clip

    def __call__(self, rng: np.random.Generator) -> float:
        if self.p < 1.0 and rng.random() > self.p:
            return 1.0

        z = rng.normal()
        if self.clip > 0:
            z = np.clip(z, -self.clip, self.clip)
        return np.exp(z * self.std)


class Uniform(Augmentation[float]):
    """Uniform distribution.

    Type: `float`; returns `0.0` if not enabled.

    Args:
        p: probability of enabling this augmentation.
        lower: lower bound.
        upper: upper bound.
    """

    def __init__(
        self, p: float = 1.0, lower: float = -np.pi, upper: float = np.pi
    ) -> None:
        super().__init__()
        self.p = p
        self.lower = lower
        self.upper = upper

    def __call__(self, rng: np.random.Generator) -> float:
        if self.p < 1.0 and rng.random() > self.p:
            return 0.0

        return rng.uniform(self.lower, self.upper)
